/*****************************************************************************
 *
 * perfchart.c - Displays graphs of data.
 *
 * Copyright (c) 2004 Ben Clewett (bclewett@perfparse.org)
 * Last Modified:   2004-05-28
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <gd.h>
#include <gdfonts.h>
#include <gdfontmb.h>
#include <glib.h>
#include "common.h"
#include "dbms.h"
#include <libpp_common.h>
#include "cgi.h"
#include <config.h>
#include <gettext.h>
#include <locale.h>

// #define DEBUG

/* Don't change these please */
#define AV_DELTA_SECONDS (60)
#define AV_MAX_POINTS (24 * 60 * 60 / AV_DELTA_SECONDS)

#define CHART_AREA_BORDER  1

gdImagePtr graph = 0;

void getCGI();
void doSQL();

/* CGI in */
char sHost[76] = "";
char sService[76] = "";
char sMetric[76] = "";
char sFrom[30], sTo[30];
char sTimeRange[2];
time_t  tRelativeTime;
int  bForceScale = FALSE;
int  bAccurateScale = FALSE;
int  bPercent = FALSE;
char *sTitle = NULL;
int  bMin, bMax;
char sFileName[128];

/* Scale works on integers between 0 and 1000 */
int iScaleTop, iScaleBot, iScaleYDelta0, iScaleYDelta1, iScaleYDelta2;
double dMin, dMax, dScaleTop, dScaleBot;
int iExp;

time_t tFrom, tTo, tFromMidnight;

// Dimensions of graph:
int iWidth = 585, iHeight = 300;
const int iBorder = 30;
const int iBotBorder = 48;

// Colours:
int black, white, blue, lblue, red, grey0, grey1, grey2, yellow;

// PID of calling code:
int iParentPID = 0;

#ifdef DEBUG
FILE *debug;
#endif

// Temporary storage of values:
GMemChunk *CData;
typedef struct {
	void *next;
	void *prev;
	time_t ctime;
	float value, smooth;
	float warn, crit;
} TData;
TData *headData = NULL;

// Gaussian Distribution
#define GAUSSIAN_SIGMA_DEFAULT 7
#define GAUSSIAN_RANGE ((int)fGaussianSigma * 2)
float fGaussianSigma = GAUSSIAN_SIGMA_DEFAULT;
float *dGaussian;
void defineGaussian();
void applyGaussian() ;
float calcGaussian(TData *dataIn);
int getGaussianUnit(time_t t);

/* Statics */
int   iValueCount;
float dMaxValue, dMinValue; 
float dValueSum, dValueSquareSum;

/* Plots */
int bSelectLine, bVal, bSmooth, bWarn, bCrit, bStanDev;

time_t time2UnixTime(char *sDate);
void getValues();
void header();
void calcScale();
void drawXScale();
void drawYScalePercent();
void drawYScale();
void drawGrid();
void drawPoints();
void drawAverage();
void drawHeadings();
void getAverage();
static int yTranslate(double d);
static int yTranslateScale(int i);
static int xTranslate(time_t t);
static int yTranslatePercent(double d);
time_t getTime(char *s);
int iYear, iMonth, iDay;
char sUnit[20];
char sLabel[40];
static void addBuffer(double *iBuf, double *iBufSquare, int *iCount, time_t x0, time_t x1, double y0, double y1);
void initChartGraphics();
void getTimeRange(time_t tIn);
void getFromAndEnd();
double my_abs(double d);
void getExponent(double dValue, int *iValue, int *iExponent);
void writePIDFile();
void print_sql_failure_image(char*file, int line, const char *sql, const char *error);
void print_failure_html(FILE *f, char *sMsg);

#define TYPE_HISTEROGRAM		0
#define TYPE_AVERAGE			1
#define TYPE_INTEGRAL			2
#define TYPE_DIFFERENTIAL		3
int iType;

/* Buffers for average: (Every minute) */
double *dVal;
double *dWarn;
double *dCrit;
int *iAvCount;
double *dValSquare;



int main(int argc, char*argv[])
{

        char *argvInt[1];
	int   argcInt = 1;

	int options[] = {
		CONFIG_ITEM_ID_ERROR_LOG,
		CONFIG_ITEM_ID_ERROR_LOG_ROTATE,
		CONFIG_ITEM_ID_DB_USER,
		CONFIG_ITEM_ID_DB_PWD,
		CONFIG_ITEM_ID_DB_DATABASE,
		CONFIG_ITEM_ID_DB_HOST,
		CONFIG_ITEM_ID_DEFAULT_USER_PERMISSIONS_POLICY,
		CONFIG_ITEM_ID_DEFAULT_USER_PERMISSIONS_HOST_GROUP,
		CONFIG_ITEM_ID_DUMMY_HOSTNAME,
		CONFIG_ITEM_ID_NB
	};

	printf("Expires: Sun, 02 Jan 2000 01:01:01 GMT\n");
	printf("content-type: image/png\n\n");

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	
	#ifdef DEBUG
	debug = fopen("/tmp/chart.debug", "w");
	#endif

	argvInt[0] = argv[0];

	/* Get settings */
	config_parse_config(argcInt,argvInt,"perfchart.png\n\n","\n",options, "stderr");

	/* Get CGI variables */
	getCGI();

	/* Write tmp file reporting PID */
	writePIDFile();

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX");
	fprintf(debug, "Max %f Min %f\n", dMax, dMin);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	/* Init the chart */
	initChartGraphics();

	/* Handle SQL */
	doSQL();
	
	/* Get the bounds of the graph */
	getFromAndEnd();

	/* Get the actual data */
	getValues();

	/* Get the gussian smoothed curge: */
	if (iType != TYPE_AVERAGE && bSmooth) {
		applyGaussian();

		#ifdef DEBUG
		fprintf(debug, "Gaussian Applied\n");
		fflush(debug);
		#endif

	}

	/* Get max and min */
	if (iType == TYPE_HISTEROGRAM || iType == TYPE_AVERAGE) {
		if (!bMax) dMax = dMaxValue;
		if (!bMin) dMin = (dMinValue < 0) ? dMinValue : 0;
	} else {
		dMax = dMaxValue;
		dMin = (dMinValue < 0) ? dMinValue : 0;
	}

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "Max %f Min %f\n", dMax, dMin);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	/* Get average */
	if (iType == TYPE_AVERAGE) getAverage();

	/* Calculate the scale */
	calcScale();

	/* Draw Grid */
	drawGrid();

	/* Draw the graph */
	switch (iType) {
	case TYPE_HISTEROGRAM:
	case TYPE_INTEGRAL:
	case TYPE_DIFFERENTIAL:
		drawPoints();
		break;
	case TYPE_AVERAGE:
		drawAverage();
		break;
	}

	/* Borders */
	gdImageRectangle(graph, 0, 0, iWidth - 1, iHeight - 1, black);
	gdImageRectangle(graph, iBorder * 2, iBorder, iWidth - iBorder, iHeight - iBotBorder, black);

	/* Chop off top & bottom */
	gdImageFilledRectangle(graph, iBorder * 2, 1, iWidth - iBorder, iBorder - 1, white);
	gdImageFilledRectangle(graph, iBorder * 2, iHeight - iBotBorder + 1, iWidth - iBorder, iHeight - 2, white);

	/* Draw scale */
	drawXScale();
	if (bPercent)
		drawYScalePercent();
	else
		drawYScale();

	/* headings and key */
	drawHeadings();

	/* Splurge */
	gdImagePng(graph, stdout);

	exit(0);
}




/* To start with, get all the values into a buffer. */
/* sTo and sFrom must be set. */
void getValues()
{
	float dValue, dWarn, dCrit, dLastValue = 0;
	TData *data = NULL, *prev = NULL;
	#ifdef SHOW_STATS
	FILE *fStats;
	#endif
	time_t tTime, tLastTime = 0;

	#ifdef USE_DB_POSTGRESQL
	int numberOfRows;
	#endif

	CData = g_mem_chunk_create(TData,(int)(4096/sizeof(TData)),G_ALLOC_ONLY);

	iValueCount = 0;
	dValueSum = 0;
	dValueSquareSum = 0;

	#ifdef DEBUG
	fprintf(debug, "From '%s' to '%s'\n", sFrom, sTo);
	fflush(debug);
	#endif

	#ifdef USE_DB_MYSQL
	g_string_printf(s_SQL, "SELECT UNIX_TIMESTAMP(ctime), value, warn, critical FROM perfdata_service_bin WHERE host_name = '%s'",
		sql_escape(sHost,PERCENT_ESCAPE_CHAR));
	#elif defined USE_DB_POSTGRESQL
	g_string_printf(s_SQL, "SELECT ctime, value, warn, critical FROM perfdata_service_bin WHERE host_name = '%s'",
		sql_escape(sHost,PERCENT_ESCAPE_CHAR));
	#endif
	g_string_append_printf(s_SQL, " AND service_description = '%s'",
		sql_escape(sService,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND metric = '%s'",
		sql_escape(sMetric,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND ctime BETWEEN '%s' AND '%s' ORDER BY ctime",
		sFrom, sTo);

	/* Note:- Not using standard cursor.  That loads everything into memory, too heavey. */
	/* Using the 'Read Forward' cursor.  If edit code, please OBAY RULES about read-forward cursor. */

	/* [TODO]  When binary cursor come out from MySQL, use it instead. */

	#ifdef USE_DB_MYSQL
	if (mysql_query(&db_mysql, s_SQL->str)) 
		print_sql_failure_image(__FILE__,__LINE__, s_SQL->str, mysql_error(&db_mysql));

	query_result = mysql_use_result(&db_mysql);
	setlocale(LC_NUMERIC, "POSIX"); 

	while ((result_row = mysql_fetch_row(query_result))) {
		tTime = (time_t)iData(0);
	#elif defined USE_DB_POSTGRESQL
	query_result = PQexec(db_postgresql, s_SQL->str);
	if(PQresultStatus(query_result) == PGRES_BAD_RESPONSE || PQresultStatus(query_result) == PGRES_FATAL_ERROR) 
		print_sql_failure_image(__FILE__,__LINE__, s_SQL->str, PQresultErrorMessage(query_result));
	
	setlocale(LC_NUMERIC, "POSIX"); 
	
	numberOfRows = rows();
	for (currentRow = 0; currentRow<numberOfRows; currentRow++){
		tTime = tData(0);
	#endif
		sscanf(sData(1), "%f", &dValue);
		sscanf(sData(2), "%f", &dWarn);
		sscanf(sData(3), "%f", &dCrit);

		// Statistics:
		dValueSum += dValue;
		dValueSquareSum += dValue * dValue;

		if (dLastValue == 0 && iType == TYPE_DIFFERENTIAL) {
			// Discaud first value
			dLastValue = dValue;
			tLastTime = tTime;
			continue;
		}

		if (tTime == tLastTime && iType == TYPE_DIFFERENTIAL) {
			// Divion by zero
			dLastValue = dValue;
			tLastTime = tTime;
			continue;
		}


		/* Storage of values for later use, traditional linkes list */
		if (!(data = g_chunk_new(TData, CData))) print_failure_html(stderr, "Out of memory (1)");
		if (prev) {
			data->prev = prev;
			prev->next = data;
		} else {
			data->prev = NULL;
		}
		prev = data;
		if (!headData) headData = data;

		data->ctime = tTime;
		switch (iType) {
		case TYPE_HISTEROGRAM:
		case TYPE_AVERAGE:
			data->value = dValue;
			break;
		case TYPE_DIFFERENTIAL:
			data->value = (dValue - dLastValue) / (double)(tTime - tLastTime);
			break;
		case TYPE_INTEGRAL:
			data->value = dValueSum;
			break;
		}
		data->smooth = 0;
		data->warn = dWarn;
		data->crit = dCrit;

		/* Range */
		if (iValueCount == 0) {
			dMaxValue = dMinValue = data->value;
		} else {
			if (dMaxValue < data->value) dMaxValue = data->value;
			if (dMinValue > data->value) dMinValue = data->value;
		}

		#ifdef DEBUG
		fprintf(debug, "%d\t%f\t%f\t%f\n", (int)data->ctime, data->value, data->warn, data->crit);
		fflush(stdout);
		#endif


		dLastValue = dValue;
		tLastTime = tTime;
		iValueCount++;
	}

	setlocale(LC_NUMERIC, "");

	if (data != NULL) {
		data->next = NULL;
	}

	#ifdef DEBUG
	fprintf(debug, "Count %d\n", iValueCount);
	fflush(debug);
	#endif

	/* Write the statistics to be read by calling program */
	#ifdef SHOW_STATS
	sprintf(sFileName, "%s-STATS-%d", FILE_STATISTICS, iParentPID);
	fStats = fopen(sFileName, "w");
	fprintf(fStats, "%d %f %f %f %f\n", iValueCount, dValueSum, dValueSquareSum, dMaxValue, dMinValue);
	fflush(fStats);
	fclose(fStats);
	#endif

}


void initChartGraphics() 
{
	/* Define us an area */
	graph = gdImageCreate(iWidth, iHeight);
	if (graph == NULL) exit(0);

	/* Colours (first is the background) */
	white = gdImageColorAllocate(graph, 255, 255, 255);
    black = gdImageColorAllocate(graph, 0, 0, 0);
    blue = gdImageColorAllocate(graph, 0, 0, 255);
    lblue = gdImageColorAllocate(graph, 192, 192, 255);
	red = gdImageColorAllocate(graph, 255, 0, 0);
	yellow = gdImageColorAllocate(graph, 255, 255, 0);
	grey2 = gdImageColorAllocate(graph, 192, 192, 192);
	grey1 = gdImageColorAllocate(graph, 224, 224, 224);
	grey0 = gdImageColorAllocate(graph, 240, 240, 240);

}


void getAverage()
{

	int i;
	TData *data;

	time_t t0, t1;
	double value0 = 0, value1 = 0;
	double warn0 = 0, warn1 = 0;
	double crit0 = 0, crit1 = 0;

	/* Buffers: */
	dVal = (double *)malloc(sizeof(double) * AV_MAX_POINTS);
	dWarn = (double *)malloc(sizeof(double) * AV_MAX_POINTS);
	dCrit  = (double *)malloc(sizeof(double) * AV_MAX_POINTS);
	dValSquare = (double *)malloc(sizeof(double) * AV_MAX_POINTS);
	iAvCount = (int *)malloc(sizeof(int) * AV_MAX_POINTS);

	/* init buffers to zero.  Important for averaging! */
	for (i = 0; i < AV_MAX_POINTS; i++) {
		iAvCount[i] = 0;
		dWarn[i] = dCrit[i] = dVal[i] = dValSquare[i] = 0;
	}

	data = headData;
	t1 = 0;
	while (data) {

		/* Seconds of this day */
		t0 = t1;
		t1 = (data->ctime - tFromMidnight) % (60 * 60 * 24);

		/* value */
		value0 = value1;
		value1 = data->value;

		/* warning */
		warn0 = warn1;
		warn1 = data->warn;

		/* critical */
		crit0 = crit1;
		crit1 = data->crit;

		/* First point, or day change */
		if (t0 == 0 || t1 < t0) {
			data = data->next;
			continue;
		}

		/* Add to buffer */
		addBuffer(dVal, dValSquare, iAvCount, t0, t1, value0, value1);
		addBuffer(dWarn, NULL, NULL, t0, t1, warn0, warn1);
		addBuffer(dCrit, NULL, NULL, t0, t1, crit0, crit1);

		data = data->next;
	}

	/* Get the max */
	if (!bForceScale) {
		dMax = 0;
		dMin = 0;
		for (i = 0; i < AV_MAX_POINTS; i++) {

			if (!iAvCount[i]) continue;
			if (dVal[i]  / iAvCount[i] > dMax) dMax = dVal[i]  / iAvCount[i];
		}
	}

}


void drawHeadings()
{
	int iKeyWidth, i, yTop;
	int iTitleLeft;
	int iCount = 0;

	if (bVal) iCount++;
	if (bWarn) iCount++;
	if (bCrit) iCount++;
	if (bSmooth) iCount++;
	if (bStanDev) iCount++;

	yTop = iHeight - 23;
	iKeyWidth = iCount * 80;
	i = iWidth - iKeyWidth - iBorder;

	/* Key */
	if (bVal) {
		gdImageString(graph, gdFontSmall, i, yTop + 2, "Value", black);
		gdImageLine(graph, i + 40, yTop + 8, i + 60, yTop + 8, black);
		gdImageLine(graph, i + 40, yTop + 9, i + 60, yTop + 9, black);
		i += 80;
	}
	if (bSmooth) {
		gdImageString(graph, gdFontSmall, i, yTop + 2, "Smooth", black);
		gdImageLine(graph, i + 40, yTop + 8, i + 60, yTop + 8, blue);
		gdImageLine(graph, i + 40, yTop + 9, i + 60, yTop + 9, blue);
		i += 80;
	}
	if (bStanDev) {
		gdImageString(graph, gdFontSmall, i, yTop + 2, "S.Dev", black);
		gdImageLine(graph, i + 40, yTop + 8, i + 60, yTop + 8, lblue);
		gdImageLine(graph, i + 40, yTop + 9, i + 60, yTop + 9, lblue);
		i += 80;
	}
	if (bWarn) {
		gdImageString(graph, gdFontSmall, i, yTop + 2, "Warn", black);
		gdImageLine(graph, i + 40, yTop + 8, i + 60, yTop + 8, yellow);
		gdImageLine(graph, i + 40, yTop + 9, i + 60, yTop + 9, yellow);
		i += 80;
	}
	if (bCrit) {
		gdImageString(graph, gdFontSmall, i, yTop + 2, "Crit", black);
		gdImageLine(graph, i + 40, yTop + 8, i + 60, yTop + 8, red);
		gdImageLine(graph, i + 40, yTop + 9, i + 60, yTop + 9, red);
		i += 80;
	}
	gdImageRectangle(graph, iWidth - iBorder - iKeyWidth - 10, yTop + 1, iWidth - iBorder, yTop + 17, black);


	/* Version */
	sprintf(sLabel, "v%s", VERSION);
	gdImageString(graph, gdFontSmall, 5, iHeight - 14, sLabel, black);

	/* Unit */
	if (sUnit[0] && !bPercent) {
		switch (iType) {
		case TYPE_HISTEROGRAM:
		case TYPE_AVERAGE:
			sprintf(sLabel, "Unit: %s", sUnit);
			break;
		case TYPE_DIFFERENTIAL:
			sprintf(sLabel, "Unit: %s/sec", sUnit);
			break;
		case TYPE_INTEGRAL:
			sprintf(sLabel, "Unit: Sum %s", sUnit);
			break;
		}
		gdImageString(graph, gdFontSmall, 5, iHeight - 28, sLabel, black);
	}


	/* Title: */
	if (!sTitle[0]) {
		g_string_printf(s_SQL, "%s '%s'   %s '%s'   %s '%s'", 
			_("Host"), sHost, _("Service"), sService, _("Metric"), sMetric);
		sTitle = s_SQL->str;
	}
	iTitleLeft = iWidth / 2 - (int)((double)strlen(sTitle) * 3.5);
	gdImageString(graph, gdFontMediumBold, iTitleLeft, 6, sTitle, black);

}



void drawAverage()
{
	/* Now draw points */
	int x0 = 0, x1 = 0, y0, y1;
	int i, iLast = 0;
	double fStanDev0, fStanDev1;


	for (i = 0; i < AV_MAX_POINTS; i++) {

		/* time scale */
		x0 = x1;
		x1 = xTranslate(i * AV_DELTA_SECONDS);

		if (!i) continue;

		if (x1 < (iBorder * 2) || x0 < (iBorder * 2) || x1 > (iWidth - iBorder) || x0 > (iWidth - iBorder)) continue;

		iLast = i - 1;

		if (!iAvCount[iLast] || !iAvCount[i]) continue;

		/* display */

		/* Standard Deviation */
		if (bStanDev) {

			/* Note:- Rounding errors sometimes produce a very small negative number */
			/* instead of zero.  Eg, -0.000013 */
			/* Which fuckes the square root. */
			/* Hence the my_abs */
			/* TODO find the French for 'fuck up' */


			if (iAvCount[iLast] > 1) {
				fStanDev0 = (dValSquare[iLast] - (dVal[iLast] * dVal[iLast] / iAvCount[iLast])) / (iAvCount[iLast] - 1);
				fStanDev0 = pow(my_abs(fStanDev0), 0.5);
			} else {
				fStanDev0 = 0;
			}
			if (iAvCount[i] > 1) {
				fStanDev1 = (dValSquare[i] - (dVal[i] * dVal[i] / iAvCount[i])) / (iAvCount[i] - 1);
				fStanDev1 = pow(my_abs(fStanDev1), 0.5);
			} else {
				fStanDev1 = 0;
			}


			if (fStanDev0 || fStanDev1) {

				y0 = yTranslate((dVal[iLast] / iAvCount[iLast]) - fStanDev0);
				y1 = yTranslate((dVal[i] / iAvCount[i]) - fStanDev1);

				gdImageLine(graph, x0, y0, x1, y1, lblue);

				y0 = yTranslate((dVal[iLast] / iAvCount[iLast]) + fStanDev0);
				y1 = yTranslate((dVal[i] / iAvCount[i]) + fStanDev1);

				gdImageLine(graph, x0, y0, x1, y1, lblue);
			}

		}


		/* Warning */
		if (bWarn) {
			y0 = yTranslate(dWarn[iLast] / iAvCount[iLast]);
			y1 = yTranslate(dWarn[i] / iAvCount[i]);

			gdImageLine(graph, x0, y0, x1, y1, yellow);
		}

		/* Critical */
		if (bCrit) {
			y0 = yTranslate(dCrit[iLast] / iAvCount[iLast]);
			y1 = yTranslate(dCrit[i] / iAvCount[i]);

			gdImageLine(graph, x0, y0, x1, y1, red);
		}

	}

	/* Display value separatelly to ensure it is always on top. */

	if (bVal) {
		for (i = 0; i < AV_MAX_POINTS; i++) {

			/* time scale */
			x0 = x1;
			x1 = xTranslate(i * AV_DELTA_SECONDS);

			if (!i) continue;

			if (x1 < (iBorder * 2) || x0 < (iBorder * 2) || x1 > (iWidth - iBorder) || x0 > (iWidth - iBorder)) continue;

			iLast = i - 1;

			if (!iAvCount[iLast] || !iAvCount[i]) continue;

			/* Display */

			/* Value */
			y0 = yTranslate(dVal[iLast] / iAvCount[iLast]);
			y1 = yTranslate(dVal[i] / iAvCount[i]);

			gdImageLine(graph, x0, y0, x1, y1, black);

		}
	}

	/* You can leave out these and let the OS recover the mess afterwards. */
	/* This would slightly speed the code, but bad programming. */
	free(dVal);
	free(dWarn);
	free(dCrit);
	free(iAvCount);
	free(dValSquare);
}


double my_abs(double d)
{
	if (d < 0) d *= -1;
	return d;
}

void drawPoints()
{
	int x0 = 0, x1 = 0;
	int value0, value1;
	int average0, average1;
	int warn0, warn1;
	int crit0, crit1;
	TData *data;

	value1 = 0;
	average1 = 0;
	warn1 = 0;
	crit1 = 0;

	data = headData;
	while (data) {

		/* time scale */
		x0 = x1;
		x1 = xTranslate(data->ctime);

		/* value */
		value0 = value1;
		value1 = yTranslate(data->value);

		#ifdef DEBUG
		setlocale(LC_NUMERIC, "POSIX");
		fprintf(debug, "Plot %f -> %d,  Time %d -> %d\n", data->value, value1, (int)data->ctime, x1);
		setlocale(LC_NUMERIC, "");
		#endif

		/* Smooth curve */
		average0 = average1;
		average1 = yTranslate(data->smooth);

		/* warning */
		warn0 = warn1;
		warn1 = yTranslate(data->warn);

		/* critical */
		crit0 = crit1;
		crit1 = yTranslate(data->crit);

		/* display */
		if (x1 > (iBorder * 2) && x0 > (iBorder * 2) && x1 < (iWidth - iBorder) && x0 < (iWidth - iBorder)) {
			if (bWarn) gdImageLine(graph, x0, warn0, x1, warn1, yellow);
			if (bCrit) gdImageLine(graph, x0, crit0, x1, crit1, red);
			if (bSmooth) gdImageLine(graph, x0, average0, x1, average1, blue);
			if (bVal) gdImageLine(graph, x0, value0, x1, value1, black);
		}

		data = data->next;

	}

}




static void addBuffer(double *fBuf, double *dValSquare, int *iCount, time_t x0, time_t x1, double y0, double y1)
{
	static double c, m;
	static int x, i;
	static double y;

	/* x in seconds. y in real values. */

	/* Duplicate in source file.  Causes div by zero. */
	if (x0 == x1) return;

	/* Work in seconds. (x0 & x1)  Transfer to actual scale (i) to store. */
	/* If you want to increase presision on buffer, */
	/* entirelly up to you, alter definitions at top of program. */

	/* Linear Regression. */
	/* calculate y = mx + c, and apply to buffer. */
	m = (y1 - y0) / (double)(x1 - x0);
	c = y0 - m * (double)x0;

	for (x = x0; x < x1; x += 1) {

		y = m * x + c;

		i = x / AV_DELTA_SECONDS;

		if (i > 0 && i < AV_MAX_POINTS) {
			fBuf[i] += y;
			if (dValSquare) dValSquare[i] += y * y;
			if (iCount) iCount[i] += 1;
		}

	}

}


void getFromAndEnd()
{
	switch (iType) {
	case TYPE_AVERAGE:

		tFrom = 0;
		tTo = 60 * 60 * 24;
		break;

	case TYPE_HISTEROGRAM:
	case TYPE_INTEGRAL:
	case TYPE_DIFFERENTIAL:

		/* Get the dates into UNIX format */
		tFrom = getTime(sFrom);
		tTo = getTime(sTo);
		break;

	}

	/* Get midnight of start: */
	/* Should correct for time zone at start of plot. */
	/* time zone change during plot, unforunate... */
	strncpy(sLabel, sFrom, 10);
	sLabel[10] = 0;
	sprintf(sLabel, "%s 00:00:00", sLabel);
	tFromMidnight = getTime(sLabel);

}


void drawGrid()
{
	int i, x, y;
	time_t t;
	int p;

	/* Smallest scale */
	/* y */
	if (!bPercent) {
		i = iScaleBot;
		while (i <= iScaleTop) {
			y = yTranslateScale(i);
			gdImageLine(graph, (iBorder * 2) + 1, y, iWidth - iBorder - 1, y, grey0);
			i += iScaleYDelta2;
		}
	} else {
		if (iHeight - iBotBorder - iBorder > 300) {
			for (p = 0; p <= 100; p += 1) {
				y = yTranslatePercent((double)p / 100);
				gdImageLine(graph, (iBorder * 2) + 1, y, iWidth - iBorder - 1, y, grey0);
			}
		}
	}

	/* x */
	if ((tTo - tFrom) / 3600 < (iWidth - iBorder * 3) * 4) {
		for (t = tFromMidnight + 3600; t <= tTo; t += 3600) {
			x = xTranslate(t);
			if (x > iBorder * 2)
				gdImageLine(graph, x, iBorder, x, iHeight - iBotBorder - 1, grey0);
		}
	}

	/* Medium Scale */
	/* y */
	if (!bPercent) {
		i = iScaleBot;
		while (i <= iScaleTop) {
			y = yTranslateScale(i);
			gdImageLine(graph, (iBorder * 2) + 1, y, iWidth - iBorder - 1, y, grey1);
			i += iScaleYDelta1;
		}
	} else {
		for (p = 0; p <= 100; p += 5) {
			y = yTranslatePercent((double)p / 100);
			gdImageLine(graph, (iBorder * 2) + 1, y, iWidth - iBorder - 1, y, grey1);
		}
	}

	/* x */
	if ((tTo - tFrom) / (3600 * 6) < (iWidth - iBorder * 3) * 4) {
		for (t = tFromMidnight + (3600 * 6); t <= tTo; t += (3600 * 6)) {
			x = xTranslate(t);
			if (x > iBorder * 2)
				gdImageLine(graph, x, iBorder, x, iHeight - iBotBorder - 1, grey1);
		}
	}


	/* Bit of a botch, but a special case where time scale is less than 12 hours */
	/* redraw scale to every ten mintes. */
	if (tTo - tFrom < 3600 * 12 ) {
		for (t = tFromMidnight; t <= tTo; t += 60 * 10) {
			x = xTranslate(t);
			if (x < iBorder * 2) continue;
			if (t % 3600 == 0) 
				gdImageLine(graph, x, iBorder, x, iHeight - iBotBorder - 1, grey1);
			else
				gdImageLine(graph, x, iBorder, x, iHeight - iBotBorder - 1, grey0);
			
		}
	}

	/* Main scale */

	/* y */
	if (!bPercent) {
		i = iScaleBot;
		while (i <= iScaleTop) {
			y = yTranslateScale(i);
			gdImageLine(graph, (iBorder * 2) + 1, y, iWidth - iBorder - 1, y, grey2);
			i += iScaleYDelta0;
		}
	} else {
		for (p = 0; p <= 100; p += 10) {
			y = yTranslatePercent((double)p / 100);
			gdImageLine(graph, (iBorder * 2) + 1, y, iWidth - iBorder - 1, y, grey2);
		}
	}

	/* x */
	for (t = tFromMidnight + (3600 * 24); t <= tTo; t += (3600 * 24)) {
		x = xTranslate(t);
		gdImageLine(graph, x, iBorder, x, iHeight - iBotBorder - 1, grey2);
	}


}


void drawYScalePercent()
{
	double p, y;

	for (p = 0; p <= 1; p += 0.1) {

		y = yTranslatePercent(p);

		sprintf(sLabel, "%.0f%%", p * 100);

		gdImageString(graph, gdFontSmall, 5, y - 6, sLabel, black);

	}

}


void drawYScale()
{
	int i = 0;
	int y;
	double d;

	/* Main scale, inc units */
	i = iScaleBot;
	while (i <= iScaleTop) {

		y = yTranslateScale(i);

		#ifdef DEBUG
		fprintf(debug, "%5d -> %d\n", i, y);
		#endif

		if (y < iBorder || y > (iHeight - iBotBorder)) {
			i += iScaleYDelta0;
			continue;
		}

		/* text */
		d = (double)i * pow(10, iExp);
		if (iExp > 2 || iExp < -6) {
			sprintf(sLabel, "%3.2e", d);
		} else {
			switch (iExp) {
			case 4:
				sprintf(sLabel, "%5.0f", d);
				break;
			case 3:
				sprintf(sLabel, "%4.0f", d);
				break;
			case 2:
				sprintf(sLabel, "%3.0f", d);
				break;
			case 1:
				sprintf(sLabel, "%3.1f", d);
				break;
			case 0:
				sprintf(sLabel, "%3.2f", d);
				break;
			case -1:
				sprintf(sLabel, "%3.3f", d);
				break;
			case -2:
				sprintf(sLabel, "%4.4f", d);
				break;
			case -3:
				sprintf(sLabel, "%5.5f", d);
				break;
			case -4:
				sprintf(sLabel, "%6.6f", d);
				break;
			}
		}
		gdImageString(graph, gdFontSmall, 5, y - 6, sLabel, black);

		i += iScaleYDelta0;

	}


}

void drawXScale()
{
	int i = 0, iSpacePerDivision, iIncrement;
	int x;
	time_t t;
	struct tm *lt;

	switch (iType) {
	case TYPE_AVERAGE:

		for (t = 0; t <= 60 * 60 * 24; t += 3600 * ((iWidth > 500) ? 1 : 2)) {
			x = xTranslate(t);
			sprintf(sLabel, "%02ld", t / 3600);
			gdImageString(graph, gdFontSmall, x - 6, iHeight - iBotBorder + 2, sLabel, black); 
		}


		break;
	case TYPE_HISTEROGRAM:
	case TYPE_INTEGRAL:
	case TYPE_DIFFERENTIAL:

		/* Day labels */

		iIncrement = 3600 * 24;		/* One lable each day. */

		/* Normalise until fits in space we have */
		/* by incresing the increment until fits space */
		if (tTo - tFrom > 3600 * 24) {
			iSpacePerDivision = (iWidth - iBorder * 3) / ((tTo - tFrom) / iIncrement);
			while (iSpacePerDivision <= 30) {
				iIncrement *= 2;
				iSpacePerDivision = (iWidth - iBorder * 3) / ((tTo - tFrom) / iIncrement);
			}
		} else {
			iSpacePerDivision = -1;
		}

		if (iSpacePerDivision > 0) {

			for (t = tFromMidnight; t <= tTo; t += iIncrement) {

				x = xTranslate(t);
				if (x < iBorder * 2) continue;

				if (iSpacePerDivision >= 80) {

					lt = localtime(&t);
					sprintf(sLabel, "%04d-%02d-%02d", lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday);
					gdImageString(graph, gdFontSmall, x - 30, iHeight - iBotBorder + 12, sLabel, black); 

				} else {

					lt = localtime(&t);
					sprintf(sLabel, "%04d-", lt->tm_year + 1900);
					gdImageString(graph, gdFontSmall, x - 13, iHeight - iBotBorder + 2, sLabel, black);
					sprintf(sLabel, "%02d-%02d", lt->tm_mon + 1, lt->tm_mday);
					gdImageString(graph, gdFontSmall, x - 13, iHeight - iBotBorder + 12, sLabel, black);

				}
			} /* END for */
		
		} /* END if */


		/* Hour labels */
		if (iSpacePerDivision > 0 && iSpacePerDivision < 80) break;	/* no room for hours */

		iIncrement = 3600;					/* Try Every One hour */

		/* i = number of label points */
		/* Watch for devision by zero */
		i = (tTo - tFrom) / iIncrement;
		if (!i) break;
		iSpacePerDivision = (iWidth - iBorder * 3) / i;
		if (iSpacePerDivision < 20) {
			iIncrement = 3600 * 2;			/* Try Two hours */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}

		if (iSpacePerDivision < 20) {
			iIncrement = 3600 * 3;			/* Three hours */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 20) {
			iIncrement = 3600 * 6;			/* Six hours */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 20) {
			iIncrement = 3600 * 12;			/* Twelve hours */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 20) {
			break;							/* Too low, don't use hour labels. */
		}

		for (t = tFromMidnight; t <= tTo; t += iIncrement) {
			x = xTranslate(t);
			if (x < iBorder * 2) continue;
			sprintf(sLabel, "%02ld", ((t - tFromMidnight) % (3600 * 24)) / 3600);
			gdImageString(graph, gdFontSmall, x - 6, iHeight - iBotBorder + 2, sLabel, black); 
		}


		/* Minute labels */
		if (iSpacePerDivision > 0 && iSpacePerDivision < 80) break;	/* no room for hours */

		iIncrement = 60;					/* Try Every One minute */

		/* i = number of label points */
		/* Watch for devision by zero */
		i = (tTo - tFrom) / iIncrement;
		if (!i) break;
		iSpacePerDivision = (iWidth - iBorder * 3) / i;
		if (iSpacePerDivision < 40) {
			iIncrement = 60 * 2;			/* Try Two minutes */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}

		if (iSpacePerDivision < 40) {
			iIncrement = 60 * 3;			/* Three minutes */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 40) {
			iIncrement = 60 * 10;			/* Ten minutes */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 40) {
			iIncrement = 60 * 20;			/* Twenty minutes */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 40) {
			iIncrement = 60 * 30;			/* Thirty Minutes */
			i = (tTo - tFrom) / iIncrement;
			if (!i) break;
			iSpacePerDivision = (iWidth - iBorder * 3) / i;
		}
		if (iSpacePerDivision < 40) {
			break;							/* Too low, don't use minute labels. */
		}

		for (t = tFromMidnight; t <= tTo; t += iIncrement) {
			// if (t % 3600 == 0) continue;
			x = xTranslate(t);
			if (x < iBorder * 2) continue;
			sprintf(sLabel, "%02ld:%02ld", ((t - tFromMidnight) % (3600 * 24)) / 3600, (t / 60) % 60);
			gdImageString(graph, gdFontSmall, x - 15, iHeight - iBotBorder + 12, sLabel, black); 
		}



		break;

	}

}


static int xTranslate(time_t t)
{
	static double d;
	d = (double)t;

	d -= tFrom;
	d /= (tTo - tFrom);
	d *= iWidth - (iBorder * 3);

	d += iBorder * 2;

	return (int)d;
}


static int yTranslate(double d)
{

	d -= dMin;
	d /= (dMax - dMin);
	d *= iHeight - (iBorder + iBotBorder + CHART_AREA_BORDER * 2);
	d = iHeight - (iBotBorder + CHART_AREA_BORDER) - d;

	return (int)d;
}

static int yTranslatePercent(double d)
{
	
	d = d * (iHeight - (iBorder + iBotBorder + CHART_AREA_BORDER * 2));
	d = iHeight - (iBotBorder + CHART_AREA_BORDER) - d;

	return (int)d;
}


static int yTranslateScale(int i)
{
	static double d;
	d = (double)i * pow(10, iExp);

	d -= dMin;
	d /= (dMax - dMin);
	d *= iHeight - (iBorder + iBotBorder + CHART_AREA_BORDER * 2);
	d = iHeight - (iBotBorder + CHART_AREA_BORDER) - d;

	return (int)d;
}


void getExponent(double dValue, int *iValue, int *iExponent)
// Retruns a number from 100 to 999 and it's exponent
{
	char sRaw[20], sNo[6];
	int j, i;
	double dTmp1, dTmp2;

	sprintf(sRaw, "%3.1e", dValue);

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "'%f' -> '%s'\n", dValue, sRaw);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	if (sRaw[0] == '-') 
		i = 1;
	else 
		i = 0;
	j = 0;
	while (sRaw[i + j] != 'e') {
		sNo[j] = sRaw[i + j];
		j++;
	}
	sNo[j] = 0;
	sscanf(sNo, "%lf", &dTmp1);

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "'%f' -> '%s' -> '%s' -> '%f'\n", dValue, sRaw, sNo, dTmp1);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	/* Exponent. */
	sNo[0] = sRaw[i + 6];
	sNo[1] = sRaw[i + 7];
	sNo[2] = 0;
	sscanf(sNo, "%lf", &dTmp2);
	if (sRaw[i + 4] == '-') dTmp2 *= -1;

	*iExponent = (int)dTmp2;

	while (dTmp1 < 100 && dTmp1 != 0) {
		dTmp1 *= 10;
		*iExponent -= 1;
	}

	/* Correct for negative */
	if (sRaw[0] == '-') dTmp1 *= -1;

	*iValue = (int)dTmp1;

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "IN '%f'  -> %d Exp %d\n", dValue, *iValue, *iExponent);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

}


void calcScale()
{
	// This function get the approx. max and min near some decimal place
	// cut off.  Works out a good scale between these values with approx.
	// ten incements.  Then sets the scale to the next sacle point 
	// >= max and <= min.

	int iMax, iMin, iMaxExp, iMinExp;
	int iRange;
	int i, iCorrection;
	
	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "Max %f Min %f\n", dMax, dMin);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	// Get the range details:  (100 -> 999)
	// Sets the exponent used everywhere.
	getExponent(dMax - dMin, &iRange, &iExp);

	// Break down the max and the min:
	getExponent(dMax, &iMax, &iMaxExp);
	getExponent(dMin, &iMin, &iMinExp);

	// Get the max/min in same exponent as range:
	while (iMaxExp > iExp) {
		iMaxExp--;
		iMax *= 10;
	}
	while (iMaxExp < iExp) {
		iMaxExp++;
		iMax /= 10;
	}
	while (iMinExp > iExp) {
		iMinExp--;
		iMin *= 10;
	}
	while (iMinExp < iExp) {
		iMinExp++;
		iMin /= 10;
	}

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "iRange = %d, iMax = %d, iMin = %d, Exponent = %d\n", iRange, iMax, iMin, iExp);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	// Define a good range:
	i = iMax - iMin;
	iCorrection = 1;
	while (TRUE) {
		if (i <= 100) {
			iScaleYDelta0 = 10;
			iScaleYDelta1 = 5;
			iScaleYDelta2 = 1;
			break;
		} else if (i <= 200) {
			iScaleYDelta0 = 20;
			iScaleYDelta1 = 10;
			iScaleYDelta2 = 5;
			break;
		} else if (i <= 250) {
			iScaleYDelta0 = 25;
			iScaleYDelta1 = 10;
			iScaleYDelta2 = 5;
			break;
		} else if (i <= 500) {
			iScaleYDelta0 = 50;
			iScaleYDelta1 = 10;
			iScaleYDelta2 = 5;
			break;
		}
		i /= 10;
		iCorrection *= 10;
	}

	iScaleYDelta0 *= iCorrection;
	iScaleYDelta1 *= iCorrection;
	iScaleYDelta2 *= iCorrection;

	#ifdef DEBUG
	fprintf(debug, "iScaleYDelta0 = %d\n", iScaleYDelta0);
	fflush(debug);
	#endif

	// Set the top and bottom:
	iScaleTop = iMax - (iMax % iScaleYDelta0);
	while (iScaleTop * pow(10, iExp) < dMax) iScaleTop += iScaleYDelta0;

	iScaleBot = iMin - (iMin % iScaleYDelta0);
	while (iScaleBot * pow(10, iExp) > dMin) iScaleBot -= iScaleYDelta0;

	#ifdef DEBUG
	fprintf(debug, "iScaleTop = %d, iScaleBottom = %d\n", iScaleTop, iScaleBot);
	fflush(debug);
	#endif

	dScaleTop = (double)iScaleTop * pow(10, iExp);
	dScaleBot = (double)iScaleBot * pow(10, iExp);

	#ifdef DEBUG
	fprintf(debug, "fScaleTop = %f, fScaleBottom = %f\n", dScaleTop, dScaleBot);
	fflush(debug);
	#endif

	if (!bAccurateScale) {
		dMax = dScaleTop;
		dMin = dScaleBot;
	}


}


/***************************************
 *
 * Misk
 *
 ***************************************/


void getCGI()
{
	char *s;

	setlocale(LC_NUMERIC, "POSIX"); 

	getCGIDBVars();


	/* Graph details */
	strcpy(sHost, scgi("host"));
	strcpy(sService, scgi("service"));
	strcpy(sMetric, scgi("metric"));

	if ((s = scgi("max"))[0] > ' ') {
		sscanf(s, "%lf", &dMax);
		bMax = TRUE;
	} else {
		/* To be calculated */
		bMax = FALSE;
	}

	if ((s = scgi("min"))[0] > ' ') {
		sscanf(s, "%lf", &dMin);
		bMin = TRUE;
	} else {
		/* To be calculated */
		bMin = FALSE;
	}

	strcpy(sUnit, scgi("unit"));

	sscanf(scgi("size"), "%[^x]x%d", sLabel, &iHeight);
	iWidth = atoi(sLabel);
	if (!iWidth || !iHeight)
		iWidth = 585, iHeight = 300;

	s = scgi("output");
	if (!strcmp(s, "Histogram")) iType = TYPE_HISTEROGRAM;
	else if (!strcmp(s, "Average")) iType = TYPE_AVERAGE;
	else if (!strcmp(s, "Diff")) iType = TYPE_DIFFERENTIAL;
	else if (!strcmp(s, "Int")) iType = TYPE_INTEGRAL;

	/* Time period and type */
	strcpy(sTimeRange, scgi("timeperiod"));
	if (sTimeRange[0] == 'r') {

		/* Relative */
		tRelativeTime = atoi(scgi("reldays"));
		tRelativeTime *= 3600 * 24;
		tRelativeTime += time2UnixTime(scgi("reltime"));
		getTimeRange(tRelativeTime);

	} else {

		/* Absolute */
		strcpy(sFrom, scgi("from"));
		strcpy(sTo, scgi("to"));
		tRelativeTime = 0;

	}

	/* Scale */
	/* New: */
	bForceScale = icgi("forcescale", 0);
	/* Old: */
	s = scgi("forcemax");
	if (s && s[0]) bForceScale = TRUE;

	/* respect the max value */
	bAccurateScale = icgi("accuratescale", 0);

	/* Show values as percent */
	bPercent = icgi("percent", 0);

	/* Title */
	sTitle = strdup(scgi("title"));

	/* Lines */
	bSelectLine = icgi("select_line", 0);
	if (bSelectLine) {
		bVal = icgi("s_val", 0);
		bSmooth = icgi("s_smo", 0);
		bWarn = icgi("s_war", 0);
		bCrit = icgi("s_cri", 0);
		bStanDev = icgi("s_sta", 0);
	} else {
		bVal = TRUE;
		bSmooth = TRUE;
		bWarn = TRUE;
		bCrit = TRUE;
		bStanDev = TRUE;
	}

	/* Hide lines if type does not allow */
	if (iType == TYPE_DIFFERENTIAL || iType == TYPE_INTEGRAL) {
		bWarn = FALSE;
		bCrit = FALSE;
	}
	if (iType != TYPE_AVERAGE) {
		bStanDev = FALSE;
	} else {
		bSmooth = FALSE;
	}

	/* Gaussian Theta */
	fGaussianSigma = atof(scgi("g_sigma")) / 2;
	if (fGaussianSigma <= 0) fGaussianSigma = GAUSSIAN_SIGMA_DEFAULT;

	/* PID of calling code */
	iParentPID = icgi("pid", 0);

	setlocale(LC_NUMERIC, ""); 
}


void print_failure_html(FILE *f, char *sMsg)
{

	gdImageFilledRectangle(graph, 0, 0, iWidth, iHeight, white);

	gdImageString(graph, gdFontSmall, 10, 10, (unsigned char*)_("Error:"), black);
	gdImageString(graph, gdFontSmall, 10, 20, sMsg, black);

	/* Splurge */
	gdImagePng(graph, stdout);

	exit(EXIT_FAILURE);

}


time_t getTime(char *sYYYYMMDDhhmmss)
{
	struct tm recTime;
	static int iOffset = -1;
	time_t tOut;
	char s[5];

	if (strlen(sYYYYMMDDhhmmss) < 18) {
		int i;
		for (i = strlen(sYYYYMMDDhhmmss) + 1; i < 19; i++) sYYYYMMDDhhmmss[i] = 0;
	}

	// Year:
	s[0] = sYYYYMMDDhhmmss[0];
	s[1] = sYYYYMMDDhhmmss[1];
	s[2] = sYYYYMMDDhhmmss[2];
	s[3] = sYYYYMMDDhhmmss[3];
	s[4] = 0;
	iYear = atoi(s);
	recTime.tm_year = iYear - 1900;

	// Month:
	s[0] = sYYYYMMDDhhmmss[5];
	s[1] = sYYYYMMDDhhmmss[6];
	s[2] = 0;
	iMonth = atoi(s);
	recTime.tm_mon = iMonth - 1;

	// Month Day:
	s[0] = sYYYYMMDDhhmmss[8];
	s[1] = sYYYYMMDDhhmmss[9];
	s[2] = 0;
	iDay = atoi(s);
	recTime.tm_mday = iDay;

	// Hours:
	s[0] = sYYYYMMDDhhmmss[11];
	s[1] = sYYYYMMDDhhmmss[12];
	s[2] = 0;
	recTime.tm_hour = atoi(s);

	// Minuts:
	s[0] = sYYYYMMDDhhmmss[14];
	s[1] = sYYYYMMDDhhmmss[15];
	s[2] = 0;
	recTime.tm_min = atoi(s);

	// Seconds:
	s[0] = sYYYYMMDDhhmmss[17];
	s[1] = sYYYYMMDDhhmmss[18];
	s[2] = 0;
	recTime.tm_sec = atoi(s);
	recTime.tm_isdst = 0;

	tOut = mktime(&recTime);

	if (iOffset == -1) {
		struct tm *r;
		r = localtime(&tOut);
		r->tm_isdst = 0;
		iOffset = (int)mktime(r) - tOut;
	}

	return tOut - iOffset;
	
}




void getTimeRange(time_t tIn)
{
	/* Get a relative time into an absolute time */

	time_t t0, t1;
	struct tm *lt;

	t1 = time(NULL);
	t0 = t1 - tIn;

	lt = localtime(&t0);
	sprintf(sFrom, "%04d-%02d-%02d %02d:%02d:%02d",
		lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
		lt->tm_hour, lt->tm_min, lt->tm_sec);

	lt = localtime(&t1);
	sprintf(sTo, "%04d-%02d-%02d %02d:%02d:%02d",
		lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
		lt->tm_hour, lt->tm_min, lt->tm_sec);

}




/***************************************
 *
 * dbms
 *
 ***************************************/

void doSQL()
{
	#ifdef USE_DB_MYSQL
	libpp_mysql_set_sql_failure_callback(print_sql_failure_image);
	#elif defined USE_DB_POSTGRESQL
	libpp_postgresql_set_sql_failure_callback(print_sql_failure_image);
	#endif

	now_connect(TRUE);

}

void print_sql_failure_image(char*file, int line, const char *sql, const char *error)
{
	gchar *tmpSQL, *tmpError;
	tmpSQL = g_strdup(sql);
	tmpError = g_strdup(error);

	gdImageFilledRectangle(graph, 0, 0, iWidth, iHeight, white);

	gdImageString(graph, gdFontSmall, 10, 10, (unsigned char*)_("SQL:"), black);
	gdImageString(graph, gdFontSmall, 10, 20, tmpSQL, black);

	gdImageString(graph, gdFontSmall, 10, 40, (unsigned char*)_("Error:"), black);
	gdImageString(graph, gdFontSmall, 10, 50, tmpError, black);

	g_free(tmpSQL);
	g_free(tmpError);


	/* Splurge */
	gdImagePng(graph, stdout);

	exit(EXIT_FAILURE);

}



time_t time2UnixTime(char *sDate)
{
	char sHour[4] = "", sMinute[4] = "";
	time_t t;
	int i = 0;

	sscanf(sDate, "%[^:]:%[^:]:%d", sHour, sMinute, &i);

	t = (time_t)i;
	t += atoi(sMinute) * 60;
	t += atoi(sHour) * 3600;

	return t;

}


void defineGaussian()
{
	int i;

	dGaussian = (float *)malloc(sizeof(float) * GAUSSIAN_RANGE);

	for (i = 0; i < GAUSSIAN_RANGE; i++) {

		dGaussian[i] = (float)(
			(1 / ( fGaussianSigma * pow(2 * M_PI, 0.5))) *
			exp( - (i * i) / (2 * fGaussianSigma * fGaussianSigma) )
			);

		#ifdef DEBUG
		setlocale(LC_NUMERIC, "POSIX"); 
		fprintf(debug, "Gaussian[%d] = %f\n", i, dGaussian[i]);
		setlocale(LC_NUMERIC, ""); 
		#endif

	}

}

void applyGaussian() 
{
	TData *data;

	defineGaussian();

	data = headData;

	while (data) {

		data->smooth = calcGaussian(data);

		data = data->next;
	}

}

int getGaussianUnit(time_t t) {
// time range -> approx. pixel

	t *= iWidth - iBorder * 3;
	t /= tTo - tFrom;

	if (t < 0) t = - t;

	return t;
}

float calcGaussian(TData *dataIn) {

	static float dSumGaussian;
	static float dSum;
	static int i;
	TData *data;

	dSum = dSumGaussian = 0;

	data = dataIn;

	// Current value:
	dSumGaussian = dGaussian[0];
	dSum = data->value * dGaussian[0];

	#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	fprintf(debug, "Value in %f\n", data->value);
	setlocale(LC_NUMERIC, ""); 
	fflush(debug);
	#endif

	// Move backwards:
	data = dataIn->prev;
	while (data) {

		i = getGaussianUnit(dataIn->ctime - data->ctime);
		#ifdef DEBUG
		setlocale(LC_NUMERIC, "POSIX"); 
		fprintf(debug, "Value back = %f space = %d pixels.\n", data->value, i);
		setlocale(LC_NUMERIC, ""); 
		#endif
		if (i >= GAUSSIAN_RANGE || i < 0) break;

		dSumGaussian += dGaussian[i];
		dSum += data->value * dGaussian[i];

		data = data->prev;
	}

	// Move forwards:
	data = dataIn->next;
	while (data) {

		i = getGaussianUnit(data->ctime - dataIn->ctime);
		#ifdef DEBUG
		setlocale(LC_NUMERIC, "POSIX"); 
		fprintf(debug, "Value next = %f space = %d pixels.\n", data->value, i);
		setlocale(LC_NUMERIC, ""); 
		#endif
		if (i >= GAUSSIAN_RANGE || i < 0) break;

		dSumGaussian += dGaussian[i];
		dSum += data->value * dGaussian[i];

		data = data->next;
	}

	// Normalise & return:
	return dSum / dSumGaussian;
		
}



/* Write PID of calling code */
void writePIDFile()
{
	#ifndef SHOW_STATS

	return;

	#else

	FILE *f;

	sprintf(sFileName, "%s-PID-%d", FILE_STATISTICS, iParentPID);

	f = fopen(sFileName, "w");

	if (!f) return;

	fprintf(f, "%u\n", (unsigned int)getpid());

	fflush(f);

	fclose(f);

	#endif

}



