/*****************************************************************************
 *
 * check_perfparse_version.c - Checks the version of PerfParse database is
 * correct.  Uses Nagios check protocol
 *
 * Copyright (c) 2005 Ben Clewett (bclewett@perfparse.org)
 * Last Modified:   2005-01-07
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <glib.h>
#include <sys/stat.h>
#include <libpp_common.h>
#include "db_macro_actions.h"
#include "common.h"
#include "dbms.h"
#include "convert.h"
#include <config.h>
#include <gettext.h>
#include <locale.h>


/* Internal functions */
void print_sql_failure(char *file, int num, const char *sql, const char *error);

void my_on_exit();

/***************************************
 *
 * Main
 *
 ***************************************/


#define NAGIOS_EXIT_OK		0
#define NAGIOS_EXIT_WARN	1
#define NAGIOS_EXIT_CRIT	2
#define NAGIOS_EXIT_UNKNOWN	3

int main(int argc, char **argv)
{
	int options[] = {
		CONFIG_ITEM_ID_DB_DATABASE,
		CONFIG_ITEM_ID_DB_HOST,
		CONFIG_ITEM_ID_DB_PWD,
		CONFIG_ITEM_ID_DB_USER,
	};

	switch(config_parse_config(argc,argv,_("check_perfparse_version [options]\n\n"),"\n",options, LOCALSTATEDIR "/perfparse-db-tool.log")) {
		case CONFIG_OPTION_HELP: config_print_help(stdout, _("check_perfparse_version [options]\n\n"), "\n", options); exit(EXIT_SUCCESS);
		case CONFIG_OPTION_VERSION: config_print_version(stdout, _("check_perfparse_version [options]\n\n"), "\n"); exit(EXIT_SUCCESS);
		case CONFIG_OPTION_SHOW_CONFIG: config_show_config(stdout, _("check_perfparse_version [options]\n\n"), "\n", options); exit(EXIT_SUCCESS);
		case CONFIG_OPTION_NONE: break;
	}

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	/* Register the error options */
	#ifdef USE_DB_MYSQL
	libpp_mysql_set_sql_failure_callback(print_sql_failure);
	#elif defined USE_DB_POSTGRESQL
	libpp_postgresql_set_sql_failure_callback(print_sql_failure);
	#endif

	/* Connect to MySQL: */
	now_connect(TRUE);

	/* Register the on_exit: */
	atexit(my_on_exit);

	/* Is database verion ok? */
	if (is_db_incorrect_version()) {
		
		setlocale(LC_NUMERIC, "POSIX");

		printf("%s %s %.2f %s %s | pp_ver=%.2f true_ver=%s\n", 
			_("ERROR Database schema version is not correct."),
			_("Current version"), getVersion(),
			_("Required version"), DB_VERSION_CHAR,
			getVersion(), DB_VERSION_CHAR);

		setlocale(LC_NUMERIC, "");
	
		exit(NAGIOS_EXIT_CRIT);

	}


	printf("%s %s. | pp_ver=%s true_ver=%s\n",
		_("OK Perfparse Database Version Correct:"), 
		DB_VERSION_CHAR, DB_VERSION_CHAR, DB_VERSION_CHAR);

	exit(NAGIOS_EXIT_OK);

}


void my_on_exit()
{
	#ifdef USE_DB_MYSQL
	mysql_close(&db_mysql);
	#elif defined USE_DB_POSTGRESQL
	PQfinish(db_postgresql);
	#endif
}



void print_sql_failure(char*file,int num, const char *sql, const char *error)
{

        printf(_("An error occured with the SQL:\n\n  \"%s\"\n\nFailure Message:\n\n  \"%s\"\n\n"),
                sql, error);

        exit(NAGIOS_EXIT_UNKNOWN);

}


