/*****************************************************************************
 *
 * convert.c - Automatic conversion of database for new versions.
 *
 * Copyright (c) 2004 Ben Clewett (perfparse@clewett.org.uk)
 * Last Modified:   2004-05-28
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>
#include <libpp_common.h>
#include "common.h"
#include "dbms.h"
#include <config.h>
#include "convert.h"
#include "db_macro_actions.h"
#include <gettext.h>


/* Domain to use: */
#define SERVICE "service"

void addIdFields();
void checkTables();
void addTableHostGroup(void);
void addTableDeletePolicy(void);
void addTableRegistry(void);
void addTablePrefs(void);
void addTableGraphs(void);
void addTableBinSummaryDelPolicy();
void addTableBinSummary();
void addTableBinSummaryHeader();
void addTableBinSummaryData();
void addTableUsers();
void addTableRawSummary();
void addTableRawSummaryData();


void install_database(void) {
	fprintf(stderr, _("Not implemented yet. perfparse-db-tool --update is implemented\n"));
}

/* Main call */
void autoConvert(void)
{

	double version = 0;

	if (version < DB_VERSION) {
		printf(_("Auto-convert database up to version %s...\n\n"),DB_VERSION_CHAR);
		printf(_("Starting...\n\n"));
		sleep(1);
	} else {
		printf(_("Coversion not required.\n"));
		exit(0);
	}

	/* Start transaction */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("SET AUTOCOMMIT=0");
		query("START TRANSACTION");
	}
	
	/* Get the version from the database */
	version = getVersion();

	if (version < 0.15) {
		printf(_("Warning, this stage may take some time:\n"));
		addField("perfdata_service_raw", "nagios_status", "TINYINT DEFAULT NULL");
	}

	if (version < 0.16) {
		addField("perfdata_graphs", "refresh_time", "INT NOT NULL DEFAULT 0");
	}

	if (version < 0.17) {
		/* Add and populate the new ID fields: */
		addIdFields();
	}
	
	if (version < 0.18) {
		/* Drop the old summary tables: */
		query("DROP TABLE IF EXISTS perfdata_summary");
		query("DROP TABLE IF EXISTS perfdata_summary_data");
		/* Also drop new ones, error in last version Ben :) */
		query("DROP TABLE IF EXISTS perfdata_bin_summary_data");
	}

	/* Always check the tables */
	/* Create where missing */
	checkTables();

	/* Set to current: */
	version = DB_VERSION;
	setVersion(version);

	/* commit changes */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("COMMIT");
		query("SET AUTOCOMMIT=1");
	}

	printf(_("Conversions complete.\n"));
}



void addIdFields() 
/* 
 * Adds the host_id, service_id and metric_id
 * and populates to the correct values.
 *
 * Magor conversion Christmas 2004, Ben.
 *
 */
{
	int iNext;
	#ifdef USE_DB_POSTGRESQL
	int numberOfRows;
	#endif

	/* 1. Hosts */

	if (addField("perfdata_host", "host_id", "INT")) {
		query_no_return("ALTER TABLE perfdata_host ADD UNIQUE(host_id)");
	}

	iNext = query_int("SELECT MAX(host_id) FROM perfdata_host");
	iNext++;

	query("SELECT host_name, host_id FROM perfdata_host");

	#ifdef USE_DB_MYSQL
	while ((result_row = mysql_fetch_row(query_result))) {
	#elif defined USE_DB_POSTGRESQL
	numberOfRows = rows();
	for (currentRow=0;currentRow<numberOfRows;currentRow++){
	#endif	
	
		if (iData(1) == 0) {
			g_string_printf(s_SQL, "UPDATE perfdata_host SET host_id = %d WHERE host_name = '%s'",
				iNext, sql_escape(sData(0), PERCENT_ESCAPE_CHAR));
			query_no_return(s_SQL->str);
			iNext++;
		}

	}


	/* 2. Services */

	if (addField("perfdata_service", "service_id", "INT")) {
		query_no_return("ALTER TABLE perfdata_service ADD UNIQUE(service_id)");
	}

	iNext = query_int("SELECT MAX(service_id) FROM perfdata_service");
	iNext++;

	query("SELECT host_name, service_description, service_id FROM perfdata_service");

	#ifdef USE_DB_MYSQL
	while ((result_row = mysql_fetch_row(query_result))) {
	#elif defined USE_DB_POSTGRESQL
	numberOfRows = rows();
	for (currentRow=0;currentRow<numberOfRows;currentRow++){
	#endif	

		if (iData(2) == 0) {
			g_string_printf(s_SQL, "UPDATE perfdata_service SET service_id = %d WHERE service_description = '%s'",
				iNext, sql_escape(sData(1), PERCENT_ESCAPE_CHAR));
			g_string_append_printf(s_SQL, " AND host_name = '%s'",
				sql_escape(sData(0), PERCENT_ESCAPE_CHAR));
			query_no_return(s_SQL->str);
			iNext++;
		}

	}

	
	/* 3. Metrics */

	if (addField("perfdata_service_metric", "metric_id", "INT")) {
		query_no_return("ALTER TABLE perfdata_service_metric ADD UNIQUE(metric_id)");
	}

	iNext = query_int("SELECT MAX(metric_id) FROM perfdata_service_metric");
	iNext++;

	query("SELECT host_name, service_description, metric, metric_id FROM perfdata_service_metric");

	#ifdef USE_DB_MYSQL
	while ((result_row = mysql_fetch_row(query_result))) {
	#elif defined USE_DB_POSTGRESQL
	numberOfRows = rows();
	for (currentRow=0;currentRow<numberOfRows;currentRow++){
	#endif	

		if (iData(3) == 0) {
			g_string_printf(s_SQL, "UPDATE perfdata_service_metric SET metric_id = %d WHERE metric = '%s'",
				iNext, sql_escape(sData(2), PERCENT_ESCAPE_CHAR));
			g_string_append_printf(s_SQL, " AND service_description = '%s'",
				sql_escape(sData(1), PERCENT_ESCAPE_CHAR));
			g_string_append_printf(s_SQL, " AND host_name = '%s'",
				sql_escape(sData(0), PERCENT_ESCAPE_CHAR));
			query_no_return(s_SQL->str);
			iNext++;
		}

	}


}


void checkTables(void)
{
	/* [TODO] Make this work with PostgreSQL */

	gboolean bHostGroup = FALSE;
	gboolean bDeletePolicy = FALSE;
	gboolean bRegistry = FALSE;
	gboolean bPrefs = FALSE;
	gboolean bGraphs = FALSE;
	gboolean bUserTable = FALSE;

	/* Summary Tables, Christmas 2004 */
	gboolean bBinSummaryDelPolicy = FALSE;
	gboolean bBinSummary = FALSE;
	gboolean bBinSummaryHeader = FALSE;
	gboolean bBinSummaryData = FALSE;

	/* Summary raw tables */
	gboolean bRawSummary = FALSE;
	gboolean bRawSummaryData = FALSE;
	
	#ifdef USE_DB_MYSQL
	query("SHOW TABLES");
	while ((result_row = mysql_fetch_row(query_result))) {

		if (!strcmp(sData(0), "perfdata_host_group")) bHostGroup = TRUE;
		if (!strcmp(sData(0), "perfdata_delete_policy")) bDeletePolicy = TRUE;
		if (!strcmp(sData(0), "perfdata_registry")) bRegistry = TRUE;
		if (!strcmp(sData(0), "perfdata_prefs")) bPrefs = TRUE;
		if (!strcmp(sData(0), "perfdata_graphs")) bGraphs = TRUE;
		if (!strcmp(sData(0), "perfdata_users")) bUserTable = TRUE;

		if (!strcmp(sData(0), "perfdata_bin_summary_del_policy")) bBinSummaryDelPolicy = TRUE;
		if (!strcmp(sData(0), "perfdata_bin_summary")) bBinSummary = TRUE;
		if (!strcmp(sData(0), "perfdata_bin_summary_header")) bBinSummaryHeader = TRUE;
		if (!strcmp(sData(0), "perfdata_bin_summary_data")) bBinSummaryData = TRUE;
		if (!strcmp(sData(0), "perfdata_raw_summary")) bRawSummary = TRUE;
		if (!strcmp(sData(0), "perfdata_raw_summary_data")) bRawSummaryData = TRUE;

		
	}

	if (!bHostGroup) addTableHostGroup();
	if (!bDeletePolicy) addTableDeletePolicy();
	if (!bRegistry) addTableRegistry();
	if (!bPrefs) addTablePrefs();
	if (!bGraphs) addTableGraphs();
	if (!bUserTable) addTableUsers();

	if (!bBinSummaryDelPolicy) addTableBinSummaryDelPolicy();
	if (!bBinSummary) addTableBinSummary();
	if (!bBinSummaryHeader) addTableBinSummaryHeader();
	if (!bBinSummaryData) addTableBinSummaryData();
	if (!bRawSummary) addTableRawSummary();
	if (!bRawSummaryData) addTableRawSummaryData();

	#endif

}

void addTableBinSummaryDelPolicy() {

	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_bin_summary_del_policy ("
        		"	policy_name VARCHAR(20) NOT NULL PRIMARY KEY,"
        		"       delete_policy INT DEFAULT NULL"
			"		) type=InnoDB");
	query_no_return(s_SQL->str);
}


void addTableBinSummary() {

	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_bin_summary ("
			"	frequency INT NOT NULL,"
			"	PRIMARY KEY (frequency),"
			"	notes TEXT NOT NULL DEFAULT '',"
			"	is_deleted TINYINT NOT NULL DEFAULT 0,"
			"	default_delete INT NOT NULL DEFAULT 1000"
			"		) type=InnoDB");
	query_no_return(s_SQL->str);

}


void addTableBinSummaryHeader() {

	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_bin_summary_header ("
			"	metric_id INT NOT NULL,"
			"	frequency INT NOT NULL,"
			"	PRIMARY KEY (metric_id, frequency),"
			"	FOREIGN KEY (metric_id)"
        		"		REFERENCES perfdata_service_metric (metric_id)"
        		"		ON DELETE RESTRICT,"
			"	INDEX frequency_idx (frequency),"
			"	FOREIGN KEY (frequency)"
        		"		REFERENCES perfdata_bin_summary (frequency)"
        		"		ON DELETE RESTRICT,"
			"	last_data INTEGER,"
			"	bin_delete_policy INTEGER DEFAULT NULL,"
			"	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,"
			"	bin_delete_policy_type ENUM('template','user') NOT NULL,"
			"	is_deleted TINYINT DEFAULT 0"
			"		) type=InnoDB");
	query_no_return(s_SQL->str);

}


void addTableBinSummaryData() {
	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_bin_summary_data ("
			"	metric_id INT NOT NULL,"
			"	frequency INT NOT NULL,"
			"	ctime DATETIME NOT NULL,"
			"	PRIMARY KEY (metric_id, frequency, ctime),"
			"	FOREIGN KEY (metric_id, frequency)"
        		"		REFERENCES perfdata_bin_summary_header (metric_id, frequency)"
        		"		ON DELETE RESTRICT,"
			"	val_count INTEGER NOT NULL DEFAULT 0,"
			"	sum_val DOUBLE NOT NULL,"
			"	sum_square_val DOUBLE NOT NULL,"
			"	max_val DOUBLE NOT NULL,"
			"	min_val DOUBLE NOT NULL,"
			"	max_warn DOUBLE NOT NULL,"
			"	min_warn DOUBLE NOT NULL,"
			"	max_critical DOUBLE NOT NULL,"
			"	min_critical DOUBLE NOT NULL"
			"		) type=InnoDB");
	query_no_return(s_SQL->str);

}



void addTablePrefs(void) {
	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_prefs ("
			"		user VARCHAR(20) NOT NULL PRIMARY KEY,"
			"		group_name VARCHAR(75),"
			"		INDEX host_ix0 (group_name),"
			"		FOREIGN KEY (group_name) REFERENCES perfdata_host_group (group_name) ON DELETE SET NULL"
			"		) type=InnoDB");
	query_no_return(s_SQL->str);
}

void addTableGraphs(void) {
	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_graphs ("
			"		id INT PRIMARY KEY NOT NULL AUTO_INCREMENT,"
			"		group_name VARCHAR(75),"
			"		INDEX host_ix0 (group_name),"
			"		FOREIGN KEY (group_name) REFERENCES perfdata_host_group (group_name) ON DELETE SET NULL,"
			"		title VARCHAR(80),"
			"		descript TEXT,"
			"		url TEXT"
			"		) type=InnoDB");
	query_no_return(s_SQL->str);
}

void addTableRegistry(void) {
	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_registry ("
			"        host VARCHAR(75) DEFAULT NULL,"
			"        rkey VARCHAR(80) NOT NULL,"
			"        PRIMARY KEY (host, rkey),"
			"        rvalue TEXT,"
			"        ctime DATETIME NOT NULL "
			"        ) type=InnoDB");
	query_no_return(s_SQL->str);
}

void addTableUsers(void) {
	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_users ("
			"		user_name VARCHAR(20) NOT NULL,"
			"		PRIMARY KEY (user_name),"
			"		password VARCHAR(20),"
			"		group_name VARCHAR(75)"
			"			REFERENCES perfdata_host_group (group_name)"
			"			ON DELETE SET NULL,"
			"		raw_show_normal TINYINT NOT NULL DEFAULT 1,"
			"		raw_show_warning TINYINT NOT NULL DEFAULT 1,"
			"		raw_show_critical TINYINT NOT NULL DEFAULT 1,"
			"		raw_show_unknown TINYINT NOT NULL DEFAULT 1,"
			"		raw_refresh_time INT NOT NULL DEFAULT 0"
			"        ) type=InnoDB");
	query_no_return(s_SQL->str);
}



void addTableHostGroup(void)
{


	/* Add new host group table and link */
	query_no_return("CREATE TABLE IF NOT EXISTS perfdata_host_group ( group_name VARCHAR(75) PRIMARY KEY) type=InnoDB");
	addField("perfdata_host", "group_name", "VARCHAR(75)");
	query_no_return("ALTER TABLE perfdata_host ADD INDEX host_ix0 (group_name)");
	query_no_return("ALTER TABLE perfdata_host ADD FOREIGN KEY (group_name) REFERENCES perfdata_host_group(group_name) ON DELETE SET NULL");



}

void addTableDeletePolicy(void) {
	/* Field added in 0.06 and removed in 0.08 */
		dropField("perfdata_service_raw", "delete_policy");

	/* Fields added in 0.11 */
	query_no_return("CREATE TABLE IF NOT EXISTS perfdata_delete_policy ( "
			"policy_name VARCHAR(20) NOT NULL PRIMARY KEY, "
			"delete_policy INT DEFAULT NULL "
			") TYPE=InnoDB");

	/* table perfdata_host */
	addField("perfdata_host","raw_delete_policy","INTEGER DEFAULT NULL");
	addField("perfdata_host","raw_delete_policy_name","VARCHAR(20) DEFAULT NULL");
	addField("perfdata_host","raw_delete_policy_type","ENUM('user','template') NOT NULL");
	query_no_return("ALTER TABLE perfdata_host ADD INDEX host_ix1 (raw_delete_policy_name)");
	query_no_return("ALTER TABLE perfdata_host ADD "
			"FOREIGN KEY (raw_delete_policy_name) "
			"REFERENCES perfdata_delete_policy(policy_name) "
			"ON DELETE SET NULL");

	addField("perfdata_host","bin_delete_policy","INTEGER DEFAULT NULL");
	addField("perfdata_host","bin_delete_policy_name","VARCHAR(20) DEFAULT NULL");
	addField("perfdata_host","bin_delete_policy_type","ENUM('user','template') NOT NULL");
	query_no_return("ALTER TABLE perfdata_host ADD INDEX host_ix2 (bin_delete_policy_name)");
	query_no_return("ALTER TABLE perfdata_host ADD "
			"FOREIGN KEY (bin_delete_policy_name) "
			"REFERENCES perfdata_delete_policy(policy_name) "
			"ON DELETE SET NULL");
	query_no_return("UPDATE perfdata_host SET raw_delete_policy_type='template'");
	query_no_return("UPDATE perfdata_host SET bin_delete_policy_type='template'");

	addField("perfdata_host","is_deleted","TINYINT DEFAULT 0");

	/* table perfdata_service */
	addField("perfdata_service","raw_delete_policy","INTEGER DEFAULT NULL");
	addField("perfdata_service","raw_delete_policy_name","VARCHAR(20) DEFAULT NULL");
	addField("perfdata_service","raw_delete_policy_type","ENUM('template','user','host') NOT NULL");
	query_no_return("ALTER TABLE perfdata_service ADD INDEX service_ix2 (raw_delete_policy_name)");
	query_no_return("ALTER TABLE perfdata_service ADD "
			"FOREIGN KEY (raw_delete_policy_name) "
			"REFERENCES perfdata_delete_policy(policy_name) "
			"ON DELETE SET NULL");
	query_no_return("UPDATE perfdata_service SET raw_delete_policy=delete_policy, raw_delete_policy_type='user'");
	dropField("perfdata_service","delete_policy");

	addField("perfdata_service","is_deleted","TINYINT DEFAULT 0");

	/* table perfdata_service_metric */
	addField("perfdata_service_metric","bin_delete_policy","INTEGER DEFAULT NULL");
	addField("perfdata_service_metric","bin_delete_policy_name","VARCHAR(20) DEFAULT NULL");
	addField("perfdata_service_metric","bin_delete_policy_type","ENUM('template','user','host') NOT NULL");
	query_no_return("ALTER TABLE perfdata_service_metric ADD INDEX metric_ix2 (bin_delete_policy_name)");
	query_no_return("ALTER TABLE perfdata_service_metric ADD "
			"FOREIGN KEY (bin_delete_policy_name) "
			"REFERENCES perfdata_delete_policy(policy_name) "
			"ON DELETE SET NULL");
	query_no_return("UPDATE perfdata_service_metric SET bin_delete_policy=delete_policy, bin_delete_policy_type='user'");
	dropField("perfdata_service_metric","delete_policy");

	addField("perfdata_service_metric","is_deleted","TINYINT DEFAULT 0");

}	

void addTableRawSummary() {

	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_raw_summary ("
			"service_id INT NOT NULL,"
			"epoch INT NOT NULL,"
			"last_ctime INT default NULL,"
			"last_nagios_status INT default NULL,"
			"PRIMARY KEY (service_id, epoch),"
			"FOREIGN KEY (service_id)"
        	"		REFERENCES perfdata_service (service_id)"
			") TYPE=InnoDB");
	query_no_return(s_SQL->str);

}

void addTableRawSummaryData() {

	g_string_assign(s_SQL,"CREATE TABLE IF NOT EXISTS perfdata_raw_summary_data ("
			"service_id INT NOT NULL ,"
			"epoch INT NOT NULL ,"
			"epoch_start INT NOT NULL ,"
			"last_ctime INT NOT NULL ,"
			"last_nagios_status INT default NULL,"
			"count INT NOT NULL ,"
			"sum_uptime INT NOT NULL ,"
			"sum_warntime INT NOT NULL ,"
			"sum_criticaltime INT NOT NULL ,"
			"sum_undeftime INT NOT NULL ,"
			"PRIMARY KEY  (service_id,epoch,epoch_start),"
			"FOREIGN KEY (service_id, epoch) REFERENCES perfdata_raw_summary (service_id, epoch)"
			") TYPE=InnoDB");

	query_no_return(s_SQL->str);
}
