/*****************************************************************************
 *
 * log_parser.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2004 Yves Mettier (ymettier@libertysurf.fr)
 * Last Modified:   2004-08-25
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef OS_SOLARIS
#include <ieeefp.h>
#endif /* OS_SOLARIS */


#include <libnagios_perfdata_parser.h>

#if defined OS_SOLARIS
#define is_finite(x) (finite(x))
#endif
#if defined (OS_LINUX) || defined (OS_OPENBSD) || defined (OS_DARWIN)
int is_finite(double v) {
	return(finite(v));
}
#endif

log_line_t *log_line_new(void) {
	log_line_t *p;
	int i;
	p = malloc(sizeof(log_line_t));
	if(NULL==p)return(NULL);
	for(i=0; i<LOG_LINE_MACRO_NB; i++) {
		p->macro[i] = malloc(32*sizeof(char));
		p->macro_len[i] = 32;
		p->macro[i][0] = '\0';
	}
	return(p);
}

void log_line_free(log_line_t*p) {
	int i;
	for(i=0; i<LOG_LINE_MACRO_NB; i++) free(p->macro[i]);
	free(p);
}

service_state_e service_state_macro_to_int(char*ss) {
	if(!strcmp(ss,"OK")) return(SERVICE_STATE_OK);
	if(!strcmp(ss,"WARNING")) return(SERVICE_STATE_WARNING);
	if(!strcmp(ss,"CRITICAL")) return(SERVICE_STATE_CRITICAL);
	if(!strcmp(ss,"UNKNOWN")) return(SERVICE_STATE_UNKNOWN);
	return(SERVICE_STATE_UNSET);
}

perf_fields_t * 
perf_fields_new (void)
{
	perf_fields_t * data;
	data = malloc(sizeof(perf_fields_t));
	if(NULL == data) return(NULL);
	data->label = malloc(20*sizeof(char));
	if(NULL == data->label) { free(data); return(NULL); }
	data->value = malloc(20*sizeof(char));
	if(NULL == data->value) { free(data->label); free(data); return(NULL); }
	data->label_len = 20;
	data->label[0] = '\0';
	data->value_len = 20;
	data->value[0] = '\0';
	data->perf_offset = 0;
	return(data);
}

const char* perf_value_get_range(const char*str, double *start,double *end, int *is_inverted) {
	const char*p;
	char *p1;
	*is_inverted = 0;
	p = str;
	if(p[0] == '@') { *is_inverted = 1; p++; }
//	if(p[0] == '~') *start = -INFINITY;
//	else {
		*start = strtod(p,&p1);
		p = p1;
//	}
	while(p[0] && (p[0] != ':') && (p[0] != ';') && (p[0] != ' ') && (p[0] != '\t')) p++;
	if(p[0] != ':') {
		*end = *start;
		*start = 0.;
		return(p);
	}
	p++;
	if((p[0]=='\0') || (p[0] == ';') || (p[0] == ' ') || (p[0] == '\t')) {
		*end = INFINITY;
		return(p);
	}
//	if(p[0] == '~') *end = INFINITY;
//	else {
		*end = strtod(p,&p1);
		p = p1;
//	}
	while(p[0] && (p[0] != ';') && (p[0] != ' ') && (p[0] != '\t')) p++;
	return(p);
}

int perf_get_next(log_line_t *data, perf_fields_t *p) {
	char*p1,*p2,*p3;
	int l;
	while (data->macro[LOG_LINE_PERFDATA][p->perf_offset] == ' ') p->perf_offset++;
	if('\0' == data->macro[LOG_LINE_PERFDATA][p->perf_offset]) return(-1);

	p1 = data->macro[LOG_LINE_PERFDATA]+p->perf_offset;
	p->label[0] = '\0';
	l = 0;

	/* Find label */
	if(p1[0] == '\'') {
		p1++;
		p2 = p1;
		/* Check length of label */
		while(p2[0] != '\0') {
			if((p2[0] == '\\') && (p2[1] == '\'')) {
				p2++;
			} else if((p2[0] == '\\') && (p2[1] == '\\')) {
				p2++;
			} else if(p2[0] == '\'') {
				break;
			}
			p2++;
		}
		if(p2[0] == '\0') return(-2);
		if(p2[1] != '=') return(-3);
		/* Reallocate memory for label if necessary */
		l = p2-p1;
		if(l>= p->label_len) {
			p->label_len += 20;
			p->label = realloc(p->label,p->label_len*sizeof(char));
			if(NULL == p->label) return(-4);
		}
		/* Copy the label */
		p2 = p1;
		l = 0;
		while(p2[0] != '\0') {
			if((p2[0] == '\\') && (p2[1] == '\'')) {
				p2++;
			} else if((p2[0] == '\\') && (p2[1] == '\\')) {
				p2++;
			} else if(p2[0] == '\'') {
				break;
			}
			p->label[l++] = p2[0];
			p2++;
		}
		p->label[l] = '\0';
		if('\0' == p2[0]) return(-1);
		if('=' != p2[1]) return(-1);
		p1 = p2+1;
	} else {
		p2 = strchr(p1,'=');
		if(p2) {
			p3 = strchr(p1,' ');
			if(p3 && ((p3-p1)<(p2-p1))) return(-5);
			l = p2-p1;
			if((l+1)>= p->label_len) {
				p->label_len = l+1;
				p->label = realloc(p->label,p->label_len*sizeof(char));
				if(NULL == p->label) return(-6);
			}

			memcpy(p->label,p1,l);
			p->label[l] = '\0';
		} else {
			return(-7);
		}
		p1 = p2;
	}

	/* Find value string */
	p1++;


	if(p1[0] == '\'') {
		p1++;
		p2 = p1;
		/* Check length of value */
		while(p2[0] != '\0') {
			if((p2[0] == '\'') && (p2[1] == '\'')) {
				p2++;
			} else if(p2[0] == '\'') {
				break;
			}
			p2++;
		}
		/* Reallocate memory for value if necessary */
		l = p2-p1;
		if(l>= p->value_len) {
			p->value_len += 20;
			p->value = realloc(p->value,p->value_len*sizeof(char));
			if(NULL == p->value) return(-8);
		}
		/* Copy the value */
		p2 = p1;
		l = 0;
		while(p2[0] != '\0') {
			if((p2[0] == '\'') && (p2[1] == '\'')) {
				p2++;
			} else if(p2[0] == '\'') {
				break;
			}
			p->value[l++] = p2[0];
			p2++;
		}
		p->value[l] = '\0';
		if(p2[0]) p2++;
		p1 = p2;
		p->type = PERF_FIELDS_TYPE_STRING;
	} else {
		p2 = strchr(p1,' ');
		if(p2)
			l = p2-p1;
		else
			l = strlen(p1);
		if((l+1)>= p->value_len) {
			p->value_len = l+1;
			p->value = realloc(p->value,p->value_len*sizeof(char));
			if(NULL == p->value) return(-9);
		}
		memcpy(p->value,p1,l);
		p->value[l] = '\0';
		p->type = PERF_FIELDS_TYPE_PERF;
		p1+= l;
	}
	p->perf_offset = (p1 - data->macro[LOG_LINE_PERFDATA]);
	return(0);
}

int 
perf_get_first (log_line_t *data, perf_fields_t *p)
{
  for (p->perf_offset = 0; data->macro[LOG_LINE_PERFDATA][p->perf_offset] == ' '; p->perf_offset++);
  return (perf_get_next(data,p));
}

perf_value_t *perf_value_new(void) {
	perf_value_t *pv;
	pv = malloc(sizeof(perf_value_t));
	if(NULL == pv) return(NULL);
	return(pv);
}

int perfstring_to_perfvalue_but_without_service_state(const char*str, perf_value_t *p) {
	const char*p1;
	char *p2;
	/* Find value */
	p1 = str;
	p->d[PERF_VALUE_VALUE] = strtod(p1,&p2);

	/* Find unit */
	for(p1=p2; p2[0] && (p2[0] != ';'); p2++);
	if(p2-p1 > UNIT_LENGTH) return(1); /* too long */
	if((p2-p1) == 0) p->unit[0]='\0';
	else { strncpy(p->unit, p1, p2-p1); p->unit[p2-p1] = '\0'; }
	/* default values */
	p->d[PERF_VALUE_WARN_START] = 0.;	
	p->d[PERF_VALUE_CRIT_START] = 0.;	
	p->d[PERF_VALUE_WARN_END] = INFINITY;	
	p->d[PERF_VALUE_CRIT_END] = INFINITY;	
	p->d[PERF_VALUE_MIN] = NAN;	
	p->d[PERF_VALUE_MAX] = NAN;	
	p->ranges_type = 0;

	if(p2[0] != ';') return(0);
	p1 = p2+1;
	p->range = p1;
	if(p1[0] == '\0') return(0);
	/* Find warn range */
	if(p1[0] != ';') {
		int is_inverted;
		p1 = perf_value_get_range(p1, &(p->d[PERF_VALUE_WARN_START]), &(p->d[PERF_VALUE_WARN_END]), &is_inverted);
		if(is_inverted) p->ranges_type |= PERF_VALUE_RANGE_WARNING_RANGE_INSIDE;
	}
	if(p1[0] != ';') return(0);
	p1++;
	if(p1[0] == '\0') return(0);
	/* Find crit range */
	if(p1[0] != ';') {
		int is_inverted;
		p1 = perf_value_get_range(p1, &(p->d[PERF_VALUE_CRIT_START]), &(p->d[PERF_VALUE_CRIT_END]), &is_inverted);
		if(is_inverted) p->ranges_type |= PERF_VALUE_RANGE_CRITICAL_RANGE_INSIDE;
	}
	if(p1[0] != ';') return(0);
	p1++;
	/* Find min */
	if(p1[0] == '\0') return(0);
	if(p1[0] != ';')
	{
		p->d[PERF_VALUE_MIN] = strtod(p1,&p2);
		p1 = p2;
	}
	while(p1[0] && (p1[0] != ';')) p1++;
	if(p1[0] != ';') return(0);
	p1++;
	/* Find max */
	if(p1[0] == '\0') return(0);
	p->d[PERF_VALUE_MAX] = strtod(p1,&p2);
	p1 = p2;
	return(0);
}

int perfstring_to_perfvalue(const char*str, perf_value_t *p) {
	int r;
	int inside;
	r = perfstring_to_perfvalue_but_without_service_state(str,p);
	p->metric_state = SERVICE_STATE_OK;

	inside = 0;
	if((p->d[PERF_VALUE_VALUE] >= p->d[PERF_VALUE_WARN_START]) || (!is_finite(p->d[PERF_VALUE_WARN_START]))) inside = 1; 
	if(inside && is_finite(p->d[PERF_VALUE_WARN_END]) && (p->d[PERF_VALUE_VALUE] >= p->d[PERF_VALUE_WARN_END])) inside = 0; 
	if(p->ranges_type & PERF_VALUE_RANGE_WARNING_RANGE_INSIDE) inside = 1-inside;
	if(0 == inside) p->metric_state = SERVICE_STATE_WARNING;

	inside = 0;
	if((p->d[PERF_VALUE_VALUE] >= p->d[PERF_VALUE_CRIT_START]) || (!is_finite(p->d[PERF_VALUE_CRIT_START]))) inside = 1; 
	if(inside && is_finite(p->d[PERF_VALUE_CRIT_END]) && (p->d[PERF_VALUE_VALUE] >= p->d[PERF_VALUE_CRIT_END])) inside = 0; 
	if(p->ranges_type & PERF_VALUE_RANGE_CRITICAL_RANGE_INSIDE) inside = 1-inside;
	if(0 == inside) p->metric_state = SERVICE_STATE_CRITICAL;

	return(r);
}

int line_split(char *line, log_line_t *data)
{
	int i;
	char*p;
	int len;
	p=line;

	for(i=0; i<LOG_LINE_MACRO_NB; i++) {
		for(len=0; p[len] && (p[len] != '\t'); len++);
		if(p[len] == '\0') {
			if(i<(LOG_LINE_MACRO_NB-1)) {
//				fprintf(stderr,"%s:%d %d %d %s\n", __FILE__,__LINE__,i,len,p);
				return(1);
			}
		}
		data->macro[i][0] = '\0';
		if(len >= data->macro_len[i]) {
			data->macro_len[i] = len+32; /* len+1 is the min to set. With 32, we will not resize so sooner */
			data->macro[i] = realloc(data->macro[i],data->macro_len[i]*sizeof(char));
			if(NULL == data->macro[i]) return(1);
		}
		if(len>0) memcpy(data->macro[i],p,len);
		data->macro[i][len] = '\0';
		p += 1+len;

	}
	data->time = strtol(data->macro[LOG_LINE_TIMET],NULL,10);
	if((0 == data->time) && (EINVAL == errno)) {
//		fprintf(stderr,"%s:%d '%s'\n", __FILE__,__LINE__,data->macro[LOG_LINE_TIMET]); 
		return(1);
	}

	data->service_state = service_state_macro_to_int(data->macro[LOG_LINE_SERVICESTATE]);
	if(SERVICE_STATE_UNSET == data->service_state) {
//		fprintf(stderr,"%s:%d %s\n", __FILE__,__LINE__,line);
		return (1);
	}
	return (0);
}
