/*****************************************************************************
 *
 * storage_gnuplot.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2005 Yves Mettier (ymettier@libertysurf.fr)
 * Last Modified:   2005-01-06
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#define STORAGE_GNUPLOT_VERSION "0.0.1"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <libnagios_perfdata_parser.h>
#include <libpp_common.h>
#include "libpp_storage.h"

#include <config.h>
#include <gettext.h>
#include <locale.h>

static FILE*log_fd;
static time_t tm_min,tm_max;

int storage_gnuplot_store_perf_values(storage_module_t *this, const log_line_t *log_data, const perf_fields_t *perf_fields, const perf_value_t *perf_value) {

	setlocale(LC_NUMERIC, "POSIX"); 

	if((NULL != perf_value) && (perf_fields->type == PERF_FIELDS_TYPE_PERF)) {
		if((log_data->time < tm_min) || (tm_min == 0)) tm_min = log_data->time;
		if(log_data->time > tm_max)  tm_max = log_data->time;

		fprintf(log_fd, "%s \"%s/%s/%s\" %f %f %f %f %f\n", log_data->macro[LOG_LINE_TIMET],
				log_data->macro[LOG_LINE_HOSTNAME],
				log_data->macro[LOG_LINE_SERVICEDESC],
				perf_fields->label,
				perf_value->d[PERF_VALUE_VALUE],
				perf_value->d[PERF_VALUE_WARN_START],
				perf_value->d[PERF_VALUE_WARN_END],
				perf_value->d[PERF_VALUE_CRIT_START],
				perf_value->d[PERF_VALUE_CRIT_END]);
	}

	setlocale(LC_NUMERIC, ""); 

	return(0);
}

int storage_gnuplot_open_data_file(storage_module_t*this) {
	char*filename = config_get_value_as_string(CONFIG_ITEM_ID_STORAGE_GNUPLOT_OUTPUT_FILE);
	if(! filename[0]) {
		pp_log(__FILE__,__LINE__, _("Empty file name for gnuplot output"));
		this->enabled = 0;
		return(-1);
	}
	if(!strcmp("-",filename)) log_fd = stdout;
	else log_fd = fopen(filename, "w");

	if(log_fd == NULL) {
		pp_log(__FILE__,__LINE__, _("Could not create file '%s' (errno=%s)"), filename,strerror(errno));
		this->enabled = 0;
		return(-1);

	}
	fprintf(log_fd, 
			"# How to use these data with gnuplot ?\n"
			"# 1: keep only the lines you want to display, using grep for example\n"
			"# 2: save the resulting file in /tmp/data.csv for example\n"
			"# 3: launch gnuplot\n"
			"# 4: type the commends at the end of this file\n"
			"#\n");
	tm_min = 0;
	tm_max = 0;
	return(0);
}

int storage_gnuplot_close_data_file(storage_module_t*this) {
	int interval;

	fprintf(log_fd,
			"#\n"
			"# gnuplot> set terminal x11\n"
			"# gnuplot> set xdata time\n"
			"# gnuplot> set timefmt \"%%s\"\n"
		   );

	interval = (tm_max - tm_min)/20; /* 20 intervals in the graph */
	/* Intervals should be improved... */
	if(interval > 3*2592000) {
		fprintf(log_fd,
				"# gnuplot> set format x \"%%m\\n%%Y\"\n"
				"# gnuplot> set xtics 2592000\n"
			   );
	} else if(interval > 3*86400) {
		fprintf(log_fd,
				"# gnuplot> set format x \"%%d/%%m\\n%%Y\"\n"
				"# gnuplot> set xtics 86400\n"
			   );
	} else if(interval > 3*3600) {
		fprintf(log_fd,
				"# gnuplot> set format x \"%%H:%%M\\n%%d/%%m\\n%%Y\"\n"
				"# gnuplot> set xtics %d\n",(interval/1800)*3600
			   );
	} else {
		fprintf(log_fd,
				"# gnuplot> set format x \"%%H:%%M\\n%%d/%%m\\n%%Y\"\n"
				"# gnuplot> set xtics %d\n",(interval/1800)*1800
			   );
	}


	fprintf(log_fd,
			"# gnuplot> set mxtics 4\n"
			"# gnuplot> set title \"Data from perfparsed\"\n"
			"# gnuplot> plot '/tmp/data.csv' using 1:3 with linespoints title 'data', '' using 1:3 smooth bezier title 'smooth bezier', '' using 1:5 with steps title 'WARNING', '' using 1:7 with steps title 'CRITICAL'\n"
			"#\n"
			"# To ouput to a png file :\n"
			"# gnuplot> set output '/tmp/image_gnuplot.png'\n"
			"# gnuplot> set terminal png small picsize 1024 640\n"
			"# gnuplot> replot\n"
			"#\n"
		   );
	fclose(log_fd);
	return(0);
}

void storage_gnuplot_print_usage(storage_module_t *this, FILE*fd, const char*id) {
	int options[] = {
		CONFIG_ITEM_ID_STORAGE_GNUPLOT_OUTPUT_FILE,
		CONFIG_ITEM_ID_NB
	};
	if(!strcmp(id,"version")) {
		fprintf(fd, _("Storage Module : %s\nVersion: %s (built %s %s)\n"),
				this->id, this->version, __DATE__, __TIME__);
	} else if(!strcmp(id, "help")) {
		fprintf(fd, _("Storage Module : %s\n"), this->id);
		config_print_help(fd, NULL,"", options);
	} else if(!strcmp(id, "show_config")) {
		fprintf(fd, _("# Storage Module : %s\n"), this->id);
		fprintf(fd, _("# ==============================\n\n"));
		config_show_config(fd, NULL,"", options);
		fprintf(fd, "\n");
	}
}


storage_module_t* storage_module_init(void) {
	storage_module_t *sm;
	sm = storage_new("gnuplot", STORAGE_GNUPLOT_VERSION);
	if(NULL == sm) return(NULL);
	sm->store_perf_values = storage_gnuplot_store_perf_values;
	sm->connect = storage_gnuplot_open_data_file;
	sm->disconnect = storage_gnuplot_close_data_file;
	sm->print_usage = storage_gnuplot_print_usage;
	sm->enabled = 1;
	return(sm);
}

