/*****************************************************************************
 *
 * log_reader.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2004 Yves Mettier (ymettier@libertysurf.fr)
 * Last Modified:   2004-08-25
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <glib.h>
#include <libpp_common.h>
#include "log_reader.h"
#include <gettext.h>


struct sigaction sigint_sigaction_sav;
struct sigaction sigterm_sigaction_sav;
struct sigaction sigpipe_sigaction_sav;
static sigset_t sig_mask_sav;

log_fd_t **log_fd;
int log_fd_len = -1;

void log_source_sigpipe_handler(int sig) {
        if(sig == SIGPIPE) {
                close_all_log_sources();
        }
}


int log_source_type(int i) {
	if(i >= log_fd_len) return(LOG_FD_NOEXIST);
	if(log_fd[i]) return(log_fd[i]->type);
	return(LOG_FD_UNDEF);
}

void log_source_nonblock(int i) {
#ifdef PERFPARSED
	int flags;
	flags = fcntl(log_fd[i]->fd,F_GETFL);
	if(flags != -1)
	fcntl(log_fd[i]->fd,F_SETFL,flags|O_NONBLOCK);
#endif
}

int log_source_output_write(int i, const char*str1, const char*str2) {
	int l;
	if(i >= log_fd_len) return(-1);
	if(NULL == log_fd[i]) return(-1);
	if(str1) {
		l = strlen(str1);
		write(log_fd[i]->fd, str1,l);
	}
	if(str2) {
		l = strlen(str2);
		write(log_fd[i]->fd, str2,l);
	}
	return(0);	
}

int log_source_input_is_ready(int i) {
	if(i >= log_fd_len) return(0);
	if(NULL == log_fd[i]) return(0);
	return(log_fd[i]->ready);
}

void log_source_test_input(void) {
	int i;
	fd_set rfds;
	int fd_max= -1;
	struct timeval tm;
	FD_ZERO(&rfds);
	tm.tv_sec = 1;
	tm.tv_usec = 0;

	for(i=0; i<log_fd_len; i++) {
		if(log_fd[i]) {
			log_fd[i]->ready = 0;
			switch(log_fd[i]->type) {
				case LOG_FD_SOCKET_MAIN:
				case LOG_FD_FILE:
				case LOG_FD_STDIN:
				case LOG_FD_PIPE:
				case LOG_FD_CLIENT_SOCKET:
				case LOG_FD_SOCKET:
					if(log_fd[i]->fd > fd_max) fd_max = log_fd[i]->fd;
					FD_SET(log_fd[i]->fd, &rfds);
					break;
				case LOG_FD_UNDEF:
				case LOG_FD_NOEXIST:
					break;
			}
		}
	}
	if(0 < select(fd_max+1, &rfds, NULL,NULL,&tm)) {

		for(i=0; i<log_fd_len; i++) {
			if(log_fd[i]) {
				switch(log_fd[i]->type) {
					case LOG_FD_SOCKET_MAIN:
					case LOG_FD_FILE:
					case LOG_FD_STDIN:
					case LOG_FD_PIPE:
					case LOG_FD_CLIENT_SOCKET:
					case LOG_FD_SOCKET:
						if(FD_ISSET(log_fd[i]->fd, &rfds)) {
							log_fd[i]->ready = 1;
						}
				case LOG_FD_UNDEF:
				case LOG_FD_NOEXIST:
					break;
				}
			}
		}
	}
}

int read_log_file_pos_from_file(int n) {
	FILE*fd;
	char tmp[BUFSIZ];
	char*config_item_id_service_log_position_mark_file;
	char*config_item_id_service_log_position_mark_path;
	char*file_basename;
	int j;
	struct stat buf;

	log_fd[n]->file_pos = -1;
	log_fd[n]->file_last_line_hash = -1;
	if(! config_get_value_as_boolean(CONFIG_ITEM_ID_SERVICE_LOG_SAVE_POSITION)) return(0);
	config_item_id_service_log_position_mark_path = config_get_value_as_string(CONFIG_ITEM_ID_SERVICE_LOG_POSITION_MARK_PATH);
	if(!strcmp(config_item_id_service_log_position_mark_path, "")) {
		pp_log(__FILE__,__LINE__, _("No position mark path was specified. Either disable saving the position or specify a mark path."));
		exit(EXIT_FAILURE);
	}
	if(stat(config_item_id_service_log_position_mark_path, &buf)) {
		pp_log(__FILE__,__LINE__, _("Dir '%s' is not available (errno=%d; msg='%s')"), config_item_id_service_log_position_mark_path, errno, strerror(errno));
		return(-1);
	}
	if(!(buf.st_mode & S_IFDIR)) {
		pp_log(__FILE__,__LINE__, _("Dir '%s' is not a directory"), config_item_id_service_log_position_mark_path);
		return(-1);
	}

	file_basename = log_fd[n]->file_name;
	for(j=0; log_fd[n]->file_name[j]; j++) {
		if(log_fd[n]->file_name[j] == '/') file_basename = log_fd[n]->file_name+j+1;
	}
	config_item_id_service_log_position_mark_file = g_strconcat(config_item_id_service_log_position_mark_path,"/perfdata_log.",file_basename,".mark",NULL);
	fd = fopen(config_item_id_service_log_position_mark_file, "r");
	if(NULL == fd) {
		pp_log(__FILE__,__LINE__,_("Could not open '%s'"), config_item_id_service_log_position_mark_file);
		return(-1);
	}

	if(fgets(tmp,BUFSIZ,fd) ) {
		log_fd[n]->file_pos = strtol(tmp,NULL,10);
	}
	if((0 == log_fd[n]->file_pos) && ((EINVAL == errno) || (ERANGE == errno))) {
		log_fd[n]->file_pos = -1;
		fclose(fd);
		pp_log(__FILE__,__LINE__,_("Could not read log file position in '%s'"), config_item_id_service_log_position_mark_file);
		return(-1);
	}

	if(fgets(tmp,BUFSIZ,fd) ) {
		log_fd[n]->file_last_line_hash = strtoul(tmp,NULL,10);
	}
	if((0 == log_fd[n]->file_last_line_hash) && ((EINVAL == errno) || (ERANGE == errno))) {
		log_fd[n]->file_pos = -1;
		log_fd[n]->file_last_line_hash = -1;
		fclose(fd);
		pp_log(__FILE__,__LINE__,_("Could not read log line hash in '%s'"), config_item_id_service_log_position_mark_file);
		return(-1);
	}

	fclose(fd);
	return(0);
}

int write_log_file_pos_to_file(int n) {
	FILE*fd;
	struct stat buf;
	char*config_item_id_service_log_position_mark_file;
	char*file_basename;
	char*config_item_id_service_log_position_mark_path;
	int j;

	if(! config_get_value_as_boolean(CONFIG_ITEM_ID_SERVICE_LOG_SAVE_POSITION)) return(0);
	config_item_id_service_log_position_mark_path = config_get_value_as_string(CONFIG_ITEM_ID_SERVICE_LOG_POSITION_MARK_PATH);
	if(!strcmp(config_item_id_service_log_position_mark_path, "")) {
		pp_log(__FILE__,__LINE__, _("No position mark path was specified. Either disable saving the position or specify a mark path."));
		exit(EXIT_FAILURE);
	}
	if(stat(config_item_id_service_log_position_mark_path, &buf)) {
		pp_log(__FILE__,__LINE__, _("Dir '%s' is not available (errno=%d; msg='%s')"), config_item_id_service_log_position_mark_path, errno, strerror(errno));
		return(-1);
	}
	if(!(buf.st_mode & S_IFDIR)) {
		pp_log(__FILE__,__LINE__, _("Dir '%s' is not a directory"), config_item_id_service_log_position_mark_path);
		return(-1);
	}

	file_basename = log_fd[n]->file_name;
	for(j=0; log_fd[n]->file_name[j]; j++) {
		if(log_fd[n]->file_name[j] == '/') file_basename = log_fd[n]->file_name+j+1;
	}
	config_item_id_service_log_position_mark_file = g_strconcat(config_item_id_service_log_position_mark_path,"/perfdata_log.",file_basename,".mark",NULL);
	fd = fopen(config_item_id_service_log_position_mark_file, "w");
	if(NULL == fd) {
		pp_log(__FILE__,__LINE__,_("Could not open '%s' for writing"), config_item_id_service_log_position_mark_file);
		return(-1);
	}
	fprintf(fd, "%d\n%d\n", log_fd[n]->file_pos, log_fd[n]->file_last_line_hash);
	fclose(fd);
	return(0);
}

int log_source_new(void) {
	int i;
	if(0 > log_fd_len) {
		log_fd = malloc(sizeof(log_fd_t*));
		log_fd[0] = NULL;
		log_fd_len = 1;
	}
	for(i=0; i<log_fd_len; i++) {
		if(log_fd[i] == NULL) break;
	}
	if(i == log_fd_len) {
		int j;
		log_fd_len +=5;
		log_fd = realloc(log_fd, (log_fd_len)*sizeof(log_fd_t*));
		if(NULL == log_fd) { pp_log(__FILE__,__LINE__, _("Not enough memory")); exit(EXIT_FAILURE); }
		for(j=i; j<log_fd_len; j++) log_fd[j] = NULL;
	}
	log_fd[i] = malloc(sizeof(log_fd_t));
	if(NULL == log_fd[i]) { pp_log(__FILE__,__LINE__, _("Not enough memory")); exit(EXIT_FAILURE); }

	log_fd[i]->buffer = g_string_new("");
	log_fd[i]->type = LOG_FD_UNDEF;
	log_fd[i]->fd = -1;
	log_fd[i]->file_pos = -1;
	log_fd[i]->file_last_line_hash = -1;
	log_fd[i]->file_name = NULL;
	return(i);
}

int open_log_source(char*filename) {
	int i;
	i = log_source_new();
	if(!strcmp(filename,"-")) {
		log_fd[i]->type = LOG_FD_STDIN;
		log_fd[i]->fd = 0;
		log_source_nonblock(i);
	} else if(filename[0] == '>') {
		char*host = strdup(filename+1);
		char*p = strchr(host,':');
		int port;
		struct hostent server_address, *address;
		struct sockaddr_in *client_sockaddr_in;
		int optval = 1;
		struct sigaction action;

		if(NULL == p) {
			pp_log(__FILE__,__LINE__, _("Could not parse host:port server address ('%s')"), filename+1);
			free(host);
			return(-1);
		}
		p[0] = '\0';
		port = strtol(p+1,NULL,10);
		if(port <= 0) {
			pp_log(__FILE__,__LINE__, _("Could not parse host:port server address ('%s')"), filename+1);
			free(host);
			return(-1);
		}
		log_fd[i]->type = LOG_FD_CLIENT_SOCKET;


		if ((address = gethostbyname (host)) == NULL)
		{
			pp_log(__FILE__,__LINE__, _("Could not find host '%s'."), host);
			free(host);
			return(-1);
		}

		memcpy (&server_address, address, sizeof (struct hostent));

		/* Open the server socket */

		if ((log_fd[i]->fd = socket (AF_INET, SOCK_STREAM, 6)) == -1)
		{
			pp_log(__FILE__,__LINE__, _("Could not create a new socket to host '%s'."), host);
			free(host);
			log_fd[i]->fd = -1;
			return(-1);
		}

		setsockopt (log_fd[i]->fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof (int));
		/* Get infos on the server */
		if ((client_sockaddr_in = malloc (sizeof (struct sockaddr_in))) == NULL)
		{
			pp_log(__FILE__,__LINE__, _("Could not allocate memory for new socket to host '%s'."), host);
			free(host);
			log_fd[i]->fd = -1;
			return (-1);
		}

		client_sockaddr_in->sin_family = server_address.h_addrtype;
		client_sockaddr_in->sin_port = htons (port);
		memcpy ((char *) & (client_sockaddr_in->sin_addr.s_addr),
				server_address.h_addr, server_address.h_length);
		/* Connect to the server as a client */
		if (connect
				(log_fd[i]->fd, (struct sockaddr *) client_sockaddr_in,
				 sizeof (struct sockaddr_in)))
		{
			free (client_sockaddr_in);
			pp_log(__FILE__,__LINE__, _("Could not connect the socket to the server '%s'."), host);
			free(host);
			log_fd[i]->fd = -1;
			return (-1);
		}
		free (client_sockaddr_in);
		free(host);

		/* Install sigpipe handler */
		action.sa_handler = log_source_sigpipe_handler;
		sigemptyset (&(action.sa_mask));
		action.sa_flags = 0;
		sigaction (SIGPIPE, &action, NULL);

		log_source_nonblock(i);
	} else if(filename[0] == '|') {
		struct stat buf;
		if(stat(filename+1, &buf)) {
			log_fd[i]->fd = mkfifo(filename+1, 0660);
			if(0 != log_fd[i]->fd) {
				pp_log(__FILE__,__LINE__, _("Could not create fifo '%s'"), filename+1);
				return(-1);
			}
		} else if(!(buf.st_mode & S_IFIFO)) {
			pp_log(__FILE__,__LINE__, _("'%s' exists and is not a pipe"), filename+1);
			return(-1);
		}
		log_fd[i]->type = LOG_FD_PIPE;
		log_fd[i]->fd = open(filename+1, O_RDWR);
		if(log_fd[i]->fd < 0 ) {
			pp_log(__FILE__,__LINE__, _("Could not open  ('%s')"), filename+1);
			return(-1);
		}
		log_source_nonblock(i);
		if(log_fd[i]->file_name) free(log_fd[i]->file_name);
		log_fd[i]->file_name = strdup(filename+1);

	} else {
		log_fd[i]->type = LOG_FD_FILE;
		log_fd[i]->fd = open(filename, O_RDONLY);
		if(log_fd[i]->fd < 0 ) {
			pp_log(__FILE__,__LINE__, _("Could not open  ('%s')"), filename);
			return(-1);
		}
		log_fd[i]->file_name = strdup(filename);
		read_log_file_pos_from_file(i);
		if(-1 == write_log_file_pos_to_file(i)) return(-1);


		if((log_fd[i]->file_last_line_hash != -1) && (log_fd[i]->file_pos >= 0)) {
			if(-1 == lseek(log_fd[i]->fd,log_fd[i]->file_pos,SEEK_SET)) {
				lseek(log_fd[i]->fd,0,SEEK_SET);
				log_fd[i]->file_pos = 0;
			} else {
				int hash = log_fd[i]->file_last_line_hash;
				GString*buffer;

				buffer = log_reader(i, NULL);
				if(buffer) {
					if(hash != log_fd[i]->file_last_line_hash) {
						lseek(log_fd[i]->fd,0,SEEK_SET);
					} else {
#ifdef DEBUG
						pp_log(__FILE__, __LINE__, _("Start reading log file from file pos %d"), log_fd[i]->file_pos);
#endif
					}
					g_string_free(buffer,TRUE);
				}
				/* g_string_assign(log_fd[i]->buffer,""); */
			}
		}
		log_source_nonblock(i);

	}
	return(i);
}

void log_source_command(int i) {
	char*command;
	if(LOG_FD_CLIENT_SOCKET == log_source_type(i)) {
		command = g_strdup_printf("history %d %d '' '' ''\n",config_get_value_as_integer(CONFIG_ITEM_ID_HISTORY_START_TM),config_get_value_as_integer(CONFIG_ITEM_ID_HISTORY_END_TM));
		write(log_fd[i]->fd,command,strlen(command));
		g_free(command);
	}
}

int log_source_new_server(int port) {
	int i;
	int optval = 1;
	struct sockaddr_in sockname;

	i = log_source_new();
	log_fd[i]->type = LOG_FD_SOCKET_MAIN;

	if ((log_fd[i]->fd = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
		pp_log (__FILE__, __LINE__, _("Could not create socket"));
		exit (EXIT_FAILURE);
	}

	setsockopt (log_fd[i]->fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof (int));

	memset ((char *) &sockname, 0, sizeof (struct sockaddr_in));
	sockname.sin_family = AF_INET;
	sockname.sin_port = htons (port);
	sockname.sin_addr.s_addr = htonl (INADDR_ANY);

	if (bind (log_fd[i]->fd, (struct sockaddr *) &sockname, sizeof (struct sockaddr_in)) < 0) {
		pp_log (__FILE__, __LINE__, _("Failure on bind()"));
		exit (EXIT_FAILURE);
	}

	if (listen (log_fd[i]->fd, FD_SETSIZE) < 0) { /* Because we use select(), we can not accept more connections than FD_SET_SIZE */
		pp_log (__FILE__, __LINE__, _("Failure on listen()"));
		exit (EXIT_FAILURE);
	}
		log_source_nonblock(i);
	return(i);
}

int log_source_new_connected_client(int socket_id) {
	int i;
	i = log_source_new();
	log_fd[i]->type = LOG_FD_SOCKET;

	log_fd[i]->fd = accept(log_fd[socket_id]->fd, NULL,0);
		log_source_nonblock(i);
	return(i);
}

int close_log_source(int i,int shutd) {
	if(log_fd[i]) {
		if(shutd && (LOG_FD_SOCKET_MAIN == log_fd[i]->type)) shutdown(log_fd[i]->fd,2);
		if(shutd && (LOG_FD_SOCKET == log_fd[i]->type)) shutdown(log_fd[i]->fd,2);
		if(shutd && (LOG_FD_CLIENT_SOCKET == log_fd[i]->type)) shutdown(log_fd[i]->fd,2);
		if(LOG_FD_STDIN != log_fd[i]->type) close(log_fd[i]->fd);
		if(LOG_FD_FILE == log_fd[i]->type) write_log_file_pos_to_file(i);
		if(LOG_FD_PIPE == log_fd[i]->type) unlink(log_fd[i]->file_name);
		if(log_fd[i]->file_name) free(log_fd[i]->file_name);
		g_string_free(log_fd[i]->buffer, TRUE);
		free(log_fd[i]);
		log_fd[i] = NULL;
	}
	return(0);
}

void close_all_log_sources(void) {
	int i;
	for(i=0; i<log_fd_len; i++) close_log_source(i,1);
}

GString*log_reader(int i, GString*buffer) {
	int r=0;
	char*pos;
	char tmp[101];

	if(NULL == (pos = strchr(log_fd[i]->buffer->str, '\n'))) {
		if(LOG_FD_FILE == log_fd[i]->type) lseek(log_fd[i]->fd,0,SEEK_CUR);
		while(0 < (r=read(log_fd[i]->fd,tmp,10))) {
			tmp[r] = '\0';
			log_fd[i]->file_pos +=r;
			log_fd[i]->buffer = g_string_append(log_fd[i]->buffer,tmp);
			if(strchr(tmp,'\n'))  break;
			if((LOG_FD_CLIENT_SOCKET != log_fd[i]->type) && (r!= 10)) break;
		}
	}


	if(NULL == pos) pos = strchr(log_fd[i]->buffer->str, '\n');
	if(pos) {
		int lf_pos;
		guint len;
		if(NULL == buffer) buffer = g_string_new("");
		len = pos-log_fd[i]->buffer->str;
		if(len) {
			buffer = g_string_assign(buffer, log_fd[i]->buffer->str);
			buffer = g_string_truncate(buffer,len);
			log_fd[i]->buffer = g_string_erase(log_fd[i]->buffer,0,len+1);
		} else {
			buffer = g_string_assign(buffer,"");
			log_fd[i]->buffer = g_string_erase(log_fd[i]->buffer,0,1);
		}
		lf_pos = buffer->len;
		if(buffer->str[lf_pos-1] == '\r') g_string_truncate(buffer,lf_pos-1);
		if(LOG_FD_FILE == log_fd[i]->type) {
			log_fd[i]->file_pos = lseek(log_fd[i]->fd,0,SEEK_CUR)-lf_pos-log_fd[i]->buffer->len-1;
			log_fd[i]->file_last_line_hash = g_string_hash(buffer);
		}
		return(buffer);
	}
	return (NULL);

}

void sig_handler(int n) {
	char *sigstr[] = {"SIGINT","SIGTERM", N_("Unexpected signal")};
	int sigstrid;
	switch(n) {
		case SIGINT: sigstrid = 0; break;
		case SIGTERM: sigstrid = 1; break;
		default: sigstrid = 2; 
	}
	close_all_log_sources();
	pp_log(__FILE__,__LINE__, _("Caught signal (%s)"),gettext(sigstr[sigstrid]));
	exit(EXIT_FAILURE);
}

void log_reader_install_signal_handler(void) {
	struct sigaction s;
	sigset_t mask;

	sigfillset(&mask);
	s.sa_handler = sig_handler;
	s.sa_mask = mask;
	s.sa_flags = SA_RESTART;
	sigaction(SIGINT,&s,&sigint_sigaction_sav);

	sigfillset(&mask);
	s.sa_handler = sig_handler;
	s.sa_mask = mask;
	s.sa_flags = SA_RESTART;
	sigaction(SIGTERM,&s,&sigterm_sigaction_sav);

	sigfillset(&mask);
	s.sa_handler = SIG_DFL;
	s.sa_mask = mask;
	s.sa_flags = SA_RESTART;
	sigaction(SIGPIPE,&s,&sigpipe_sigaction_sav);
}

void log_reader_remove_signal_handler(void) {
	sigaction(SIGINT,&sigint_sigaction_sav,NULL);
	sigaction(SIGTERM,&sigterm_sigaction_sav,NULL);
	sigaction(SIGPIPE,&sigpipe_sigaction_sav,NULL);
}

void log_reader_mask_signals(void) {
	sigset_t mask;
	sigemptyset(&mask);
	sigaddset(&mask, SIGINT);
	sigaddset(&mask, SIGTERM);
	sigaddset(&mask, SIGPIPE);
	sigprocmask(SIG_BLOCK,&mask,&sig_mask_sav);
}

void log_reader_unmask_signals(void) {
	sigprocmask(SIG_SETMASK,&sig_mask_sav,NULL);
}
