/*****************************************************************************
 *
 * server_parser.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2004 Yves Mettier (ymettier@libertysurf.fr)
 * Last Modified:   2004-09-01
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include "log_reader.h"
#include <libnagios_perfdata_parser.h>
#include "perfparse-common.h"
#include "storage.h"
#include "storage_history.h"

#include <libpp_common.h>
#include <gettext.h>


typedef struct {
	char*cmd;
	int nb_args;
	char*description;
	void (*command_callback)(int id, const char*cmd);
} command_element_t;

void cmd_cb_exit(int id, const char*args);
void cmd_cb_help(int id, const char*args);
void cmd_cb_hello(int id, const char*args);
void cmd_cb_echo(int id, const char*args);
void cmd_cb_registry(int id, const char*args);
void cmd_cb_history(int id, const char*args);

command_element_t command_list[] = {
	{"help", 0, N_("Print help"), cmd_cb_help},
	{"hello", 0, N_("Hello test"), cmd_cb_hello},
	{"echo", -1, N_("Print the arguments in one line"), cmd_cb_echo},
	{"set_registry", 3, N_("Set the registry (args: tm key value)"), cmd_cb_registry},
	{"history", -1, N_("Get some data (args: start_tm end_tm 'host' 'metric' 'key')"), cmd_cb_history},
	{"exit", 0, N_("Exit"), cmd_cb_exit},
	{NULL,0,NULL,NULL}
};


void cmd_cb_exit(int id, const char*args) {
	log_source_output_write(id, _("User exit\n"), NULL);
	close_log_source(id,1);
}


void cmd_cb_help(int id, const char*args) {
	int i;
	char*buffer;
	int buffer_len=0;

	for(i=0; command_list[i].cmd; i++) {
			if(command_list[i].description) {
				int l = strlen(command_list[i].cmd);
				if(l>buffer_len) buffer_len = l;
			}
	}

	buffer = malloc(sizeof(char)*(buffer_len+1));
	buffer[buffer_len] = '\0';
	log_source_output_write(id,"\n", NULL);
	for(i=0; command_list[i].cmd; i++) {
			if(command_list[i].description) {
				memset(buffer, ' ',buffer_len);
				memcpy(buffer, command_list[i].cmd, strlen(command_list[i].cmd));
				if(command_list[i].nb_args) 
					log_source_output_write(id,buffer, _(" <args> : "));
				else
					log_source_output_write(id,buffer, _("        : "));
				log_source_output_write(id,gettext(command_list[i].description), "\n");
			}
	}
	free(buffer);
}

void cmd_cb_hello(int id, const char*args) {
	log_source_output_write(id,_("Hello you :)\n") ,NULL);
}


void cmd_cb_echo(int id, const char*args) {
	if(args && args[0]) {
		log_source_output_write(id,args ,"\n");
	} else {
		log_source_output_write(id,_("(No args for echo)\n"),NULL);
	}
}

void cmd_cb_registry(int id, const char*args) {
	char *str[3];
	char*str_errormessage_missing[] = {
		N_("(host name missing)\n"),
		N_("(key missing)\n"),
		N_("(value missing)\n")
	};
	char*str_errormessage_end[] = {
		N_("(unexpected end of host name)\n"),
		N_("(unexpected end of key)\n"),
		N_("(unexpected end of value)\n")
	};
	int i;
	time_t tm;
	char*s1;

	if(args && args[0]) {
		char*buf;
		buf = strdup(args);
		tm = strtol(buf,&s1,10);
		if((0 == tm) && (errno == ERANGE)) {
			log_source_output_write(id,_("set_registry: syntax error (tm missing)\n"),NULL);
			return;
		}
		for(i=0; i<3; i++) {
			while(s1[0] && s1[0] != '\'') s1++;
			if(s1[0] == '\0') {
				log_source_output_write(id,_("set_registry: syntax error "),gettext(str_errormessage_missing[i]));
				return;
			}
			s1++;
			str[i] = s1;
			while(s1[0]) {
				if(s1[0] == '\\') s1++;
				else if(s1[0] == '\'') break;
				if('\0' == s1[0]) break;
				s1++;
			}
			if(s1[0] == '\0') {
				log_source_output_write(id,_("set_registry: syntax error "),gettext(str_errormessage_end[i]));
				return;
			}
			s1[0] = '\0';
			s1++;
		}

		storage_set_registry(str[0],str[1],str[2],tm);
		free(buf);
	} else {
		log_source_output_write(id,_("set_registry <tm> '<hostname>' '<key>' '<value>'\n"),NULL);
	}
}

void cmd_cb_history(int id, const char*args) {
	char *str[3];
	char*str_errormessage_missing[] = {
		N_("(host name missing)\n"),
		N_("(metric missing)\n"),
		N_("(key missing)\n")
	};
	char*str_errormessage_end[] = {
		N_("(unexpected end of host name)\n"),
		N_("(unexpected end of metric)\n"),
		N_("(unexpected end of key)\n")
	};
	int i;
	time_t tm_start, tm_end;
	char*s1, *s2;
	GString*line = NULL;

	if(args && args[0]) {
		char*buf;
		int pid;
		buf = strdup(args);
		tm_start = strtol(buf,&s1,10);
		if((0 == tm_start) && (errno == ERANGE)) {
			log_source_output_write(id,_("history: syntax error (tm_start missing)\n"),NULL);
			return;
		}
		tm_end = strtol(s1,&s2,10);
		if((0 == tm_end) && (errno == ERANGE)) {
			log_source_output_write(id,_("history: syntax error (tm_end missing)\n"),NULL);
			return;
		}
		s1 = s2;
		for(i=0; i<3; i++) {
			while(s1[0] && s1[0] != '\'') s1++;
			if(s1[0] == '\0') {
				log_source_output_write(id,_("history: syntax error "),gettext(str_errormessage_missing[i]));
				return;
			}
			s1++;
			str[i] = s1;
			while(s1[0]) {
				if(s1[0] == '\\') s1++;
				else if(s1[0] == '\'') break;
				if('\0' == s1[0]) break;
				s1++;
			}
			if(s1[0] == '\0') {
				log_source_output_write(id,_("history: syntax error "),gettext(str_errormessage_end[i]));
				return;
			}
			s1[0] = '\0';
			s1++;
		}
		if(tm_end <= 0) tm_end += time(0);
		if(tm_start <= 0) tm_start += tm_end;
		switch((pid=fork())) {
			case -1: pp_log(__FILE__,__LINE__,_("Could not fork")); exit(EXIT_FAILURE);
			case 0:
		line = history_get_lines_first(tm_start,tm_end,str[0],str[1],str[2]);
		while(line) {
			log_source_output_write(id,line->str, "\n");
			line = history_get_lines_next(line);
		}
		close_log_source(id,1);
		exit(EXIT_SUCCESS);
			default:
		close_log_source(id,0);

		}
		
		free(buf);
	} else {
		log_source_output_write(id,_("history <tm_start (tm_start<=0 for tm_end+tm_start)> <tm_end (val<=0 for now+val)> '<hostname (empty for all)>' '<metric (empty for all)>' '<key (empty for all)>'\n"),NULL);
	}
}




void parse_command(GString*line, int id) {
	int i;
	char*cmd, *args;

	for(i=0; line->str[i]; i++) {
		if(strchr("\n\t ", line->str[i])) {
			break;
		}
	}

	args = strchr(line->str, ' ');
	if(args) {
		int l = args-line->str;
		cmd = malloc(sizeof(char)*(l+1));
		memcpy(cmd, line->str, l);
		cmd[l] = '\0';
		while(args[0] == ' ') args++;
	} else {
		cmd = strdup(line->str);
	}

	for(i=0; command_list[i].cmd; i++) {
		if(!strcmp(command_list[i].cmd, cmd)) {
			if(command_list[i].command_callback) command_list[i].command_callback(id,args);
			break;
		}
	}
	if(NULL == command_list[i].cmd) {
		log_source_output_write(id,_("Command '"),cmd);
		log_source_output_write(id,_("' unknown.\nType 'help' for help.\n"),NULL);
	}
	free(cmd);

}

