#
# Tables used for perfparse. MySQL.
#
# Use something like:
# $ cat mysql_create.sql | mysql -u nagios -p -D nagios -h localhost
#
# Structure:
#
# 'service' and 'host' relations.  
# To be dropped in future when replacement
# for serive.cfg is added.
#
# NOTE:- Please use InnoDB tables and MySQL version 4.00.11b upwards.
# (This alows transactions so that the program is crash-safe,
#  and referential actions so that bad data cannot be entered.)
#


CREATE TABLE IF NOT EXISTS perfdata_delete_policy (
	policy_name VARCHAR(20) NOT NULL PRIMARY KEY,
	delete_policy INT DEFAULT NULL
) TYPE=InnoDB;

CREATE TABLE IF NOT EXISTS perfdata_host_group (
	group_name VARCHAR(75) PRIMARY KEY
) type=InnoDB;

CREATE TABLE IF NOT EXISTS perfdata_host (
	host_id INT NOT NULL,
	UNIQUE(host_id),
	host_name VARCHAR(75) NOT NULL PRIMARY KEY,
	group_name VARCHAR(75),
	INDEX host_ix0 (group_name),
	FOREIGN KEY (group_name) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL,
	raw_delete_policy INTEGER,
	raw_delete_policy_name VARCHAR(20) DEFAULT NULL,
	raw_delete_policy_type ENUM('user','template') NOT NULL,
	INDEX host_ix1 (raw_delete_policy_name),
	FOREIGN KEY (raw_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	bin_delete_policy INTEGER,
	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,
	bin_delete_policy_type ENUM('user','template') NOT NULL,
	INDEX host_ix2 (bin_delete_policy_name),
	FOREIGN KEY (bin_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	is_deleted TINYINT DEFAULT 0
	) type=InnoDB;

CREATE TABLE IF NOT EXISTS perfdata_service (
	service_id INT NOT NULL,
	UNIQUE(service_id),
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	PRIMARY KEY (host_name, service_description),
	INDEX service_ix0 (host_name),
	FOREIGN KEY (host_name) 
		REFERENCES perfdata_host (host_name) 
		ON DELETE RESTRICT,
	last_perfdata_raw INTEGER DEFAULT NULL,
	UNIQUE INDEX service_ix1 (last_perfdata_raw),
	raw_delete_policy INTEGER DEFAULT NULL,
	raw_delete_policy_name VARCHAR(20) DEFAULT NULL,
	raw_delete_policy_type ENUM('template','user','host') NOT NULL,
	INDEX service_ix2 (raw_delete_policy_name),
	FOREIGN KEY (raw_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	is_deleted TINYINT DEFAULT 0
	) type=InnoDB;
	
#
# Tables to hold the raw perfdata for the services:
#

CREATE TABLE IF NOT EXISTS perfdata_service_raw (
	id INTEGER NOT NULL AUTO_INCREMENT PRIMARY KEY,
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	INDEX perfdata_service_raw_ix0 (host_name, service_description),
	FOREIGN KEY (host_name, service_description) 
		REFERENCES perfdata_service (host_name, service_description) 
		ON DELETE RESTRICT,
	ctime DATETIME NOT NULL,
	INDEX perfdata_service_raw_ix1 (ctime),
	txt_data VARCHAR(255),
	nagios_status TINYINT DEFAULT NULL,
	INDEX search_idx (host_name, service_description, ctime)
) type=InnoDB;

ALTER TABLE perfdata_service 
	ADD FOREIGN KEY (last_perfdata_raw) 
	REFERENCES perfdata_service_raw(id) 
	ON DELETE SET NULL;


#
# Table to hold the metrics:
#

CREATE TABLE IF NOT EXISTS perfdata_service_metric (
	metric_id INT NOT NULL,
	UNIQUE(metric_id),
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	metric VARCHAR(75) NOT NULL,
	PRIMARY KEY (host_name, service_description, metric),
	INDEX perfdata_service_metric_ix0 (host_name, service_description),
	FOREIGN KEY (host_name, service_description) 
		REFERENCES perfdata_service (host_name, service_description) 
		ON DELETE RESTRICT,
	unit VARCHAR(10),
	notes TEXT,
	value_max FLOAT,
	value_min FLOAT,
	last_perfdata_bin INTEGER,
	UNIQUE INDEX perfdata_metric_ix1 (last_perfdata_bin),
	bin_delete_policy INTEGER DEFAULT NULL,
	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,
	bin_delete_policy_type ENUM('template','user','host') NOT NULL,
	INDEX metric_ix2 (bin_delete_policy_name),
	FOREIGN KEY (bin_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	is_deleted TINYINT DEFAULT 0
	) type=InnoDB;
	
#
# Table to hold the binary perfdata
#

CREATE TABLE IF NOT EXISTS perfdata_service_bin (
	id INTEGER NOT NULL AUTO_INCREMENT PRIMARY KEY,
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	metric VARCHAR(75) NOT NULL,
	INDEX perfdata_service_bin_ix0 (host_name, service_description),
	FOREIGN KEY (host_name, service_description) 
		REFERENCES perfdata_service (host_name, service_description) 
		ON DELETE RESTRICT,
	INDEX perfdata_service_bin_ix1 (host_name, service_description, metric),
	FOREIGN KEY (host_name, service_description, metric) 
		REFERENCES perfdata_service_metric (host_name, service_description, metric) 
		ON DELETE RESTRICT,
	ctime DATETIME NOT NULL,
	INDEX perfdata_service_raw_ix1 (host_name, service_description, metric, ctime),
	value FLOAT,
	warn FLOAT,
	critical FLOAT,
	state TINYINT
	) type=InnoDB;

ALTER TABLE perfdata_service_metric 
	ADD FOREIGN KEY (last_perfdata_bin) 
	REFERENCES perfdata_service_bin(id) 
	ON DELETE SET NULL;


#
# Binary Summary Data
# Delete policy and frequency in Hours.
#

CREATE TABLE IF NOT EXISTS perfdata_bin_summary_del_policy (
	policy_name VARCHAR(20) NOT NULL PRIMARY KEY,
	delete_policy INT DEFAULT NULL
	) type=InnoDB;


CREATE TABLE IF NOT EXISTS perfdata_bin_summary (
	frequency INT NOT NULL,
	PRIMARY KEY (frequency),
	notes TEXT NOT NULL DEFAULT '',
	is_deleted TINYINT NOT NULL DEFAULT 0,
	default_delete INT NOT NULL DEFAULT 1000
	) type=InnoDB;

CREATE TABLE IF NOT EXISTS perfdata_bin_summary_header (
	metric_id INT NOT NULL,
	frequency INT NOT NULL,
	PRIMARY KEY (metric_id, frequency),
	FOREIGN KEY (metric_id) 
		REFERENCES perfdata_service_metric (metric_id) 
		ON DELETE RESTRICT,
	INDEX frequency_idx (frequency),
	FOREIGN KEY (frequency) 
		REFERENCES perfdata_bin_summary (frequency) 
		ON DELETE RESTRICT,
	last_data INTEGER,
	bin_delete_policy INTEGER DEFAULT NULL,
	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,
	bin_delete_policy_type ENUM('template','user') NOT NULL,
	is_deleted TINYINT DEFAULT 0
	) type=InnoDB;
	
CREATE TABLE IF NOT EXISTS perfdata_bin_summary_data (
	metric_id INT NOT NULL,
	frequency INT NOT NULL,
	ctime DATETIME NOT NULL,
	PRIMARY KEY (metric_id, frequency, ctime),
	FOREIGN KEY (metric_id, frequency) 
		REFERENCES perfdata_bin_summary_header (metric_id, frequency)
		ON DELETE RESTRICT,
	val_count INTEGER NOT NULL DEFAULT 0,
	sum_val DOUBLE NOT NULL,
	sum_square_val DOUBLE NOT NULL,
	max_val DOUBLE NOT NULL,
	min_val DOUBLE NOT NULL,
	max_warn DOUBLE NOT NULL,
	min_warn DOUBLE NOT NULL,
	max_critical DOUBLE NOT NULL,
	min_critical DOUBLE NOT NULL
	) type=InnoDB;

#
# Raw Summary tables
#
CREATE TABLE IF NOT EXISTS perfdata_raw_summary (
	service_id INT NOT NULL,
	epoch INT NOT NULL,
	last_ctime INT default NULL,
	last_nagios_status INT default NULL,
	PRIMARY KEY (service_id, epoch),
	FOREIGN KEY (service_id)
		REFERENCES perfdata_service (service_id)
	) TYPE=InnoDB;

CREATE TABLE IF NOT EXISTS perfdata_raw_summary_data (
	service_id INT NOT NULL ,
	epoch INT NOT NULL ,
	epoch_start INT NOT NULL ,
	last_ctime INT NOT NULL ,
	last_nagios_status INT default NULL,
	count INT NOT NULL ,
	sum_uptime INT NOT NULL ,
	sum_warntime INT NOT NULL ,
	sum_criticaltime INT NOT NULL, 
	sum_undeftime INT NOT NULL ,
	PRIMARY KEY  (service_id,epoch,epoch_start),
	FOREIGN KEY (service_id, epoch) REFERENCES perfdata_raw_summary (service_id, epoch)
	) TYPE=InnoDB;


#
# Table to hold the system state:
#

CREATE TABLE IF NOT EXISTS perfdata_state (
	domain VARCHAR(10) NOT NULL PRIMARY KEY,
	file_name TEXT NOT NULL DEFAULT '',
	line_in_file INTEGER NOT NULL DEFAULT 0,
# Checksome to ensure stil same file:
	last_line_hash VARCHAR(20) NOT NULL DEFAULT '',
# pid of this program.  NULL = exited cleanly.
	pid INTEGER NOT NULL DEFAULT 0,
	perfparse_version VARCHAR(10) NOT NULL DEFAULT '0'
	) type=InnoDB;

CREATE TABLE IF NOT EXISTS perfdata_registry (
		host VARCHAR(75) DEFAULT NULL,
		rkey VARCHAR(80) NOT NULL,
		PRIMARY KEY (host, rkey),
		rvalue TEXT,
		ctime DATETIME NOT NULL
) type=InnoDB;


#
# Table to hold users perferances,
# for instance filters used in CGI
#

CREATE TABLE IF NOT EXISTS perfdata_prefs (
	user VARCHAR(20) NOT NULL PRIMARY KEY,
	group_name VARCHAR(75),
	INDEX host_ix0 (group_name),
	FOREIGN KEY (group_name) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL
) type=InnoDB;

#
# Stored graphs:
#

CREATE TABLE IF NOT EXISTS perfdata_graphs (
	id INT PRIMARY KEY NOT NULL AUTO_INCREMENT,
	group_name VARCHAR(75),
	INDEX host_ix0 (group_name),
	FOREIGN KEY (group_name) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL,
	title VARCHAR(80),
	descript TEXT,
	url TEXT,
	refresh_time INT NOT NULL DEFAULT 0
) type=InnoDB;

#
# Remembers User Settings
#

CREATE TABLE IF NOT EXISTS perfdata_users (
	user_name VARCHAR(20) NOT NULL,
	PRIMARY KEY (user_name),
	password VARCHAR(20),
	group_name VARCHAR(75) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL,
	raw_show_normal TINYINT NOT NULL DEFAULT 1,
	raw_show_warning TINYINT NOT NULL DEFAULT 1,
	raw_show_critical TINYINT NOT NULL DEFAULT 1,
	raw_show_unknown TINYINT NOT NULL DEFAULT 1,
	raw_refresh_time INT NOT NULL DEFAULT 0
) type=InnoDB;
